import math
import re
import hashlib
from pc23 import xrange
from .base import AttributeValueVirtual, AttributeValueNull, AttributeValueString
from ..exceptions import OrmValueError, OrmAttributeTypeMistake
from ..utils import ipaddr_version, ipv4mask_2_cidr


class ValuePassword(AttributeValueString):
    def try_password(self, password):
        value = self.value
        if value is None or isinstance(value, AttributeValueNull):
            return password is None or isinstance(password, AttributeValueNull)
        encryption = self.c.encryption
        if encryption == 'plain':
            trying = password
        elif encryption == 'md5':
            trying = hashlib.md5(password.encode('utf-8')).hexdigest().lower()
        elif encryption == 'sha1':
            trying = hashlib.sha1(password.encode('utf-8')).hexdigest().lower()
        elif encryption == 'sha224':
            trying = hashlib.sha224(password.encode('utf-8')).hexdigest().lower()
        elif encryption == 'sha256':
            trying = hashlib.sha256(password.encode('utf-8')).hexdigest().lower()
        elif encryption == 'sha384':
            trying = hashlib.sha384(password.encode('utf-8')).hexdigest().lower()
        elif encryption == 'sha512':
            trying = hashlib.sha512(password.encode('utf-8')).hexdigest().lower()
        else:
            trying = None
        return value.lower() == trying


class ValueYearMonth(AttributeValueVirtual):
    basecls = False
    check_modified = True

    def __init__(self, value):
        super(ValueYearMonth, self).__init__()
        self._value = value

    def __repr__(self):
        if self.null:
            return "<NULL>"
        y, m = self._decompose()
        c = self._get_c()
        if c is None:
            return self._value
        r = (str("%04d" % y), str("%02d" % m)) if c.year_first else (str("%04d" % m), str("%02d" % y))
        return self.c.separator.join(r)

    def base(self, value):
        return None if value is None or isinstance(value, AttributeValueNull) else float(value)

    @property
    def value(self):
        return None if self._value is None else float(self._value)

    @property
    def string(self):
        if self.null:
            return ""
        return self.__repr__()

    def _get_c(self):
        return getattr(self.instance.__class__, self.model_k) if self.model_k is not None else None

    def _decompose(self):
        y, m = divmod(self._value, 1)
        return int(round(y)), int(round(m*100))

    def get_value(self):
        return self

    @property
    def year(self):
        return None if self._value is None else self._decompose()[0]

    @year.setter
    def year(self, value):
        if value is None or isinstance(value, AttributeValueNull):
            value = 0
        y, m = self._decompose() if not self.null else (0, 0)
        y = value
        self._value = float(y) + (float(m) / 100)

    @property
    def month(self):
        return None if self._value is None else self._decompose()[0]

    @month.setter
    def month(self, value):
        if value is None or isinstance(value, AttributeValueNull):
            value = 0
        y, m = self._decompose() if not self.null else (0, 0)
        m = value
        self._value = float(y) + (float(m) / 100)

    @property
    def year_month(self):
        return None if self.null else self._decompose()

    @property
    def null(self):
        return self._value is None

    @null.setter
    def null(self, value):
        if bool(value):
            c = self._get_c()
            if c is not None and not c.nullable:
                raise OrmValueError("cannot be NULL because this attribute is not nullable!")
            self._value = None
        else:
            self._value = 0.0

    def set_value(self, instance, key, value):
        before_ = self._value
        if isinstance(value, ValueYearMonth):
            if value.null:
                self.null = True
            else:
                self.null = False
                self.year, self.month = value.year_month
            return self._value, before_
        if value is None or isinstance(value, AttributeValueNull):
            c = self._get_c()
            if c is not None and not c.nullable:
                raise OrmValueError("cannot be NULL because this attribute is not nullable!")
            self._value = None
            return self._value, before_
        if isinstance(value, float):
            self._value = value
            return self._value, before_
        if not isinstance(value, (list, tuple)) or len(value) != 2:
            raise OrmValueError(
                "YearMonth requires that value to be set using [Year, Month] format or using '.month' and"
                " '.year' properties!"
            )
        self.year = value[0]
        self.month = value[1]
        return self._value, before_


class ValuePhone(AttributeValueVirtual):
    basecls = False
    check_modified = True

    def __init__(self, value):
        super(ValuePhone, self).__init__()
        self._value = value

    def __repr__(self):
        if self.null:
            return "<NULL>"
        return self._format_phone()

    def base(self, value):
        return None if value is None or isinstance(value, AttributeValueNull) else str(value)

    @property
    def value(self):
        return self._value

    @property
    def string(self):
        if self.null:
            return ""
        return self.__repr__()

    def _get_c(self):
        return getattr(self.instance.__class__, self.model_k) if self.model_k is not None else None

    def _format_phone(self):
        if self._value is None:
            return ""
        c = self._get_c()
        return c.format_phone(*self._decompose()) if c is not None else self._value

    def _decompose(self):
        if self._value is None:
            return None
        if "," not in self._value:
            return self._value, None
        return self._value.split(",")

    def get_value(self):
        return self

    @property
    def null(self):
        return self._value is None

    @null.setter
    def null(self, value):
        if bool(value):
            c = self._get_c()
            if c is not None and not c.nullable:
                raise OrmValueError("cannot be NULL because this attribute is not nullable!")
            self._value = None
        else:
            self._value = 0.0

    def clear(self):
        c = self._get_c()
        self._value = None if c is not None and c.nullable else ""

    @property
    def phone(self):
        if self._value is None:
            return None
        return self._decompose()[0]

    @property
    def ext(self):
        if self._value is None:
            return None
        return self._decompose()[1]

    @phone.setter
    def phone(self, value):
        c = self._get_c()
        if c is None:
            return
        phone, ext = self._decompose() if self._value else (None, None)
        phone = self.c.do_store_local_prefix(re.sub("[^0-9]", "", str(value)))
        r = phone
        if ext:
            r += "," + str(ext)
        self._value = r

    @ext.setter
    def ext(self, value):
        phone, ext = self._decompose() if self._value else (None, None)
        if not value:
            self._value = phone
            return
        ext = re.sub("[^0-9]", "", str(value))
        if not phone:
            phone = ""
        self._value = "%s,%s" % (str(phone), str(ext))

    def set_value(self, instance, key, value):
        before_ = self._value
        if isinstance(value, ValuePhone):
            self.phone, self.ext = value.phone, value.ext
            return self._value, before_
        if value is None:
            self.clear()
            return self._value, before_
        if isinstance(value, (list, tuple)):
            if len(value) > 2:
                raise OrmValueError("Phone attribute can be set to (str) or (list|tuple) of length 1 or 2!")
            if len(value) == 2:
                self.phone, self.ext = value
            else:
                self.phone, self.ext = value[0], None
            return self._value, before_
        elif isinstance(value, str):
            if ',' in value:
                pc = value.split(',')
                if len(pc) > 2:
                    raise OrmValueError(
                        "(str) value for Phone attribute when set might contain comma as delimiter"
                        " for ext.phone, but only one! More than one been given: '%s'!" % str(value)
                    )
                if len(pc) == 2:
                    self.phone, self.ext = pc
                else:
                    self.phone, self.ext = pc[0], None
                return self._value, before_
            else:
                self.phone, self.ext = value, None
                return self._value, before_
        elif isinstance(value, int):
            self.phone, self.ext = str(value), None
            return self._value, before_
        else:
            raise OrmValueError("value for Phone can be type of (str|list|tuple) only!")


class ValueIPRequesties(AttributeValueVirtual):
    basecls = False
    check_modified = True

    def __init__(self, value):
        super(ValueIPRequesties, self).__init__()
        self._ipaddr = None
        self._ipmask = None
        self._set(value)

    def __repr__(self):
        if self.null:
            return "<NULL>"
        return self._str()

    def base(self, value):
        return None if self.null else self.value

    @property
    def value(self):
        return self
        # return None if self.null else self._str()

    @property
    def string(self):
        return self._str()

    def _get_c(self):
        return getattr(self.instance.__class__, self.model_k) if self.model_k is not None else None

    def get_value(self):
        return self

    @property
    def null(self):
        return self._ipaddr is None

    @null.setter
    def null(self, value):
        if bool(value):
            c = self._get_c()
            if c is not None and not c.nullable:
                raise OrmValueError("cannot be NULL because this attribute is not nullable!")
            self._ipaddr = None
        else:
            self.ipaddr, self.ipmask = '0.0.0.0', 0

    def _str(self):
        if self.null:
            return ""
        return self._repr()

    def _repr(self):
        if self.null:
            return None
        c = self._get_c()
        if c is None:
            return "%s/%s" % (str(self._ipaddr), str(self._ipmask))
        r = [str(self._ipaddr), ]
        if not self._is_host_addr() or not c.hide_host_mask:
            r.append(str(self._ipmask))
        return "/".join(r)

    def _is_host_addr(self):
        if self.null:
            return True
        ipvers = ipaddr_version(self._ipaddr)
        return self._ipmask == 32 if ipvers == 4 else (self._ipmask == 128 or self._ipmask == 64)

    @staticmethod
    def _verify_ipaddr(value):
        if value is None or isinstance(value, AttributeValueNull):
            return None
        ipvers = ipaddr_version(value)
        if not ipvers:
            raise OrmValueError("IPRequesties requires correct IP address to be given, '%s' got!" % str(value))
        return str(value)

    def _verify_ipmask(self, value, ipaddr=None):
        ipaddr_ = ipaddr or self._ipaddr
        if ipaddr_ is None or isinstance(ipaddr_, AttributeValueNull):
            return None
        ipvers = ipaddr_version(ipaddr_)
        c = self._get_c()
        if (value is None or isinstance(value, AttributeValueNull)) and ipaddr_ is not None:
            if c is not None:
                value = c.default_v6prefix if ipvers == 6 else c.default_v4mask
            else:
                value = 0
        if not isinstance(value, (str, int)):
            raise OrmValueError("IPRequesties requires that IP mask be given using (int|str) type!")
        if ipvers == 6 and not isinstance(value, int) and not str(value).isdigit():
            raise OrmValueError("IPRequesties requires that IPv6 prefix be a number!")
        if ipvers == 4:
            ipprefix = ipv4mask_2_cidr(value) \
                if not isinstance(value, int) and not str(value).isdigit() \
                else int(value)
            if ipprefix < 0:
                raise OrmValueError("IPRequesties requires that IPv4 mask cidr be >= 0!")
            if ipprefix > 32:
                raise OrmValueError("IPRequesties requires that IPv4 mask cidr be <= 32!")
        else:
            ipprefix = int(value)
            if ipprefix < 0:
                raise OrmValueError("IPRequesties requires that IPv6 prefix be >= 0!")
            if ipprefix > 128:
                raise OrmValueError("IPRequesties requires that IPv6 prefix be <= 128!")
        return int(ipprefix)

    def _verify(self, ipaddr, ipmask):
        if ipaddr is None or isinstance(ipaddr, AttributeValueNull):
            return None, None
        if not isinstance(ipaddr, str):
            raise OrmValueError("IPRequesties requires IP address to be a string, '%s' got!" % str(type(ipaddr)))
        ipaddr_ = self._verify_ipaddr(ipaddr)
        ipmask_ = self._verify_ipmask(ipmask, ipaddr_)
        return ipaddr_, ipmask_

    def _set(self, value):
        if value is None or isinstance(value, AttributeValueNull):
            self.null = True
            return
        elif isinstance(value, str):
            if '/' not in value:
                ipaddr = value
                ipmask = None
            elif value.count('/') > 1:
                raise OrmValueError("IPRequesties requires 'ipaddr/ipmask' format when list|tuple is used!")
            else:
                ipaddr, ipmask = value.split('/')
        elif not isinstance(value, (list, tuple)):
            ipaddr = value
            ipmask = None
        elif len(value) > 2:
            raise OrmValueError("IPRequesties requires [ipaddr, ipmask] format when list|tuple is used!")
        else:
            ipaddr, ipmask = value
        if ipaddr is None or isinstance(ipaddr, AttributeValueNull):
            self.null = True
            return
        if not isinstance(ipaddr, str):
            raise OrmValueError("IPRequesties requires that IP address to be a (str) type!")
        ipaddr_, ipmask_ = self._verify(ipaddr, ipmask)
        if ipaddr_ is None or isinstance(ipaddr_, AttributeValueNull):
            self.null = True
            return
        self._ipaddr, self._ipmask = ipaddr_, ipmask_

    @property
    def ipaddr(self):
        return self._ipaddr

    @property
    def ipmask(self):
        return self._ipmask

    @property
    def ipreq(self):
        return (self._ipaddr, self._ipmask) if not self.null else (None, None)

    @ipaddr.setter
    def ipaddr(self, value):
        if value is None or isinstance(value, AttributeValueNull):
            self.null = True
            return
        self._ipaddr = self._verify_ipaddr(value)

    @ipmask.setter
    def ipmask(self, value):
        if self.null:
            raise OrmValueError("cannot set IP mask or prefix without IP address defined!")
        self._ipmask = self._verify_ipmask(value)

    def set_value(self, instance, key, value):
        before_ = self._repr()
        if isinstance(value, ValueIPRequesties):
            if value.null:
                self.null = True
            else:
                self.null = False
                self.ipaddr, self.ipmask = value.ipreq
            return self._repr(), before_
        if value is None or isinstance(value, AttributeValueNull):
            c = self._get_c()
            if c is not None and not c.nullable:
                raise OrmValueError("cannot be NULL because this attribute is not nullable!")
            self.null = True
            return self._repr(), before_
        self._set(value)
        return self._repr(), before_


class ValueGeometryAbstract(AttributeValueVirtual):
    basecls = False
    check_modified = True

    def __init__(self, value):
        super(ValueGeometryAbstract, self).__init__()
        self._value = value
        self._geometry = list()
        self._length = 0
        self._separator = 'x'
        self._space_edging = True

    def __repr__(self):
        if self.null:
            return "<NULL>"
        return self._value

    def base(self, value):
        return None if value is None or isinstance(value, AttributeValueNull) else str(value)

    def _set_geom_length(self, value):
        self._length = value
        self._reset_geometry()

    @property
    def value(self):
        return self._value

    @property
    def string(self):
        if self.null:
            return ""
        return self.__repr__()

    def _get_c(self):
        return getattr(self.instance.__class__, self.model_k) if self.model_k is not None else None

    def get_value(self):
        return self

    def _cast_geom_value(self, value):
        c = self._get_c()
        if c is None:
            return str(value)
        if isinstance(value, int):
            return str(value) if not c.decimals else str(("%0."+str(c.decimals)+"f") % float(value))
        if isinstance(value, str):
            try:
                value = float(value.strip())
            except ValueError:
                raise OrmValueError("geometry value must be a type of (int) or (float)!")
        if isinstance(value, float):
            return str(int(round(value))) if not c.decimals else str(("%0."+str(c.decimals)+"f") % float(value))
        raise OrmValueError("geometry value must be a type of (int) or (float)!")

    def _get_geom_part(self, i):
        if i >= len(self._geometry):
            raise OrmAttributeTypeMistake("requested geometry exceeds existing format!")
        if self.null:
            return None
        value = self._geometry[i]
        c = self._get_c()
        if c is None:
            return value
        if c.decimals == 0:
            return int(value)
        else:
            return float(value)

    def _set_geom_part(self, i, value):
        if i >= len(self._geometry):
            raise OrmAttributeTypeMistake("requested geometry exceeds existing format!")
        before_ = self._value
        if self.null:
            self._clear_to_zero()
        self._geometry[i] = self._cast_geom_value(value)
        self._compose_value_from_geometry()
        self.instance.__check_modified__(self.model_k, (self._value, before_))

    def _reset_geometry(self):
        c = self._get_c()
        v = '0' if c is None else ('0' if c.decimals == 0 else str(("%0."+str(c.decimals)+"f") % float(0.0)))
        self._geometry = list()
        for i in xrange(self._length):
            self._geometry.append(v)

    def _compose_value_from_geometry(self):
        sep = " %s " % str(self._separator) if self._space_edging else str(self._separator)
        self._value = sep.join(self._geometry)

    def _set_from_geometry(self, value):
        self._reset_geometry()
        if value is None:
            self._value = None
            return
        if not isinstance(value, (list, tuple)):
            raise OrmValueError("geometry must be set using (list|tuple) type!")
        if len(value) == 0:
            value = list('0' * self._length)
        if len(value) > self._length:
            value = value[:self._length]
        for i in xrange(len(value)):
            self._geometry[i] = self._cast_geom_value(value[i])
        self._compose_value_from_geometry()

    def _clear_to_zero(self):
        self._set_from_geometry([])

    @property
    def null(self):
        return self._value is None

    @null.setter
    def null(self, value):
        if bool(value):
            c = self._get_c()
            if c is not None and not c.nullable:
                raise OrmValueError("cannot be NULL because this attribute is not nullable!")
            self._value = None
        else:
            self._clear_to_zero()

    def clear(self):
        c = self._get_c()
        if c is not None and c.nullable:
            self._set_from_geometry(None)
        else:
            self._clear_to_zero()

    @property
    def geometry(self):
        return self._geometry if self._value is not None else None

    @geometry.setter
    def geometry(self, value):
        if not isinstance(value, (list, tuple)):
            raise OrmValueError("geometry must be set using (list|tuple) type!")
        self._set_from_geometry(value)

    def _decompose_raw(self, value):
        value = value.replace('x', '|').replace(',', '|').replace(self._separator, '|')
        str_parts = value.split('|')
        geo_parts = list()
        for p in str_parts:
            geo_parts.append(self._cast_geom_value(re.sub("[^0-9.|\-]", "", str(p))))
        return geo_parts

    def set_value(self, instance, key, value):
        before_ = self._value
        if isinstance(value, ValueGeometryAbstract):
            value = value._value
        if value is None or isinstance(value, AttributeValueNull):
            self.null = True
            return self._value, before_
        if isinstance(value, (list, tuple)):
            self._set_from_geometry(value)
            return self._value, before_
        elif isinstance(value, str):
            if value.strip() == "":
                self._clear_to_zero()
                return self._value, before_
            self._set_from_geometry(self._decompose_raw(value))
            return self._value, before_
        else:
            raise OrmValueError("value for geometry can be type of (str|list|tuple) only!")


class ValueDimensions2D(ValueGeometryAbstract):
    def __init__(self, value):
        super(ValueDimensions2D, self).__init__(value)
        self._set_geom_length(2)

    @property
    def width(self):
        return self._get_geom_part(0)

    @property
    def height(self):
        return self._get_geom_part(1)

    @width.setter
    def width(self, value):
        self._set_geom_part(0, value)

    @height.setter
    def height(self, value):
        self._set_geom_part(1, value)


class ValueDimensions3D(ValueGeometryAbstract):
    def __init__(self, value):
        super(ValueDimensions3D, self).__init__(value)
        self._set_geom_length(3)

    @property
    def width(self):
        return self._get_geom_part(0)

    @property
    def height(self):
        return self._get_geom_part(1)

    @property
    def depth(self):
        return self._get_geom_part(2)

    @width.setter
    def width(self, value):
        self._set_geom_part(0, value)

    @height.setter
    def height(self, value):
        self._set_geom_part(1, value)

    @depth.setter
    def depth(self, value):
        self._set_geom_part(2, value)


class ValuePoint(ValueGeometryAbstract):
    def __init__(self, value):
        super(ValuePoint, self).__init__(value)
        self._separator = ','
        self._space_edging = False
        self._set_geom_length(2)

    @property
    def x(self):
        return self._get_geom_part(0)

    @property
    def y(self):
        return self._get_geom_part(1)

    @x.setter
    def x(self, value):
        self._set_geom_part(0, value)

    @y.setter
    def y(self, value):
        self._set_geom_part(1, value)


class ValueBox(ValueGeometryAbstract):
    def __init__(self, value):
        super(ValueBox, self).__init__(value)
        self._separator = ','
        self._space_edging = False
        self._set_geom_length(4)

    @property
    def top_left(self):
        return self._get_geom_part(0)

    @property
    def top_right(self):
        return self._get_geom_part(1)

    @property
    def bottom_right(self):
        return self._get_geom_part(2)

    @property
    def bottom_left(self):
        return self._get_geom_part(3)

    @top_left.setter
    def top_left(self, value):
        self._set_geom_part(0, value)

    @top_right.setter
    def top_right(self, value):
        self._set_geom_part(1, value)

    @bottom_right.setter
    def bottom_right(self, value):
        self._set_geom_part(2, value)

    @bottom_left.setter
    def bottom_left(self, value):
        self._set_geom_part(3, value)

    @property
    def width(self):
        if self.null:
            return None
        return int(round(abs(self.top_right - self.top_left) + abs(self.bottom_right - self.bottom_left)) / 2)

    @property
    def height(self):
        if self.null:
            return None
        return int(round(abs(self.bottom_left - self.top_left) + abs(self.bottom_right - self.top_right)) / 2)

    @width.setter
    def width(self, value):
        if self.null:
            self.top_left = 0
            self.bottom_left = 0
        self.top_right = self.top_left + value
        self.bottom_right = self.bottom_left + value

    @height.setter
    def height(self, value):
        if self.null:
            self.top_left = 0
            self.bottom_left = 0
        self.bottom_left = self.top_left + value
        self.bottom_right = self.top_right + value

    @property
    def corners(self):
        if self.null:
            return None
        return self._geometry


class ValueVector(ValueGeometryAbstract):
    def __init__(self, value):
        super(ValueVector, self).__init__(value)
        self._separator = ','
        self._space_edging = False
        self._set_geom_length(4)

    @property
    def x1(self):
        return self._get_geom_part(0)

    @property
    def y1(self):
        return self._get_geom_part(1)

    @property
    def x2(self):
        return self._get_geom_part(2)

    @property
    def y2(self):
        return self._get_geom_part(3)

    @property
    def length(self):
        if self.null:
            return None
        a = abs(self.x2 - self.x1)
        b = abs(self.y2 - self.y1)
        return math.sqrt(a**2 + b**2)

    @property
    def start(self):
        if self.null:
            return None
        return self.x1, self.y1

    @property
    def end(self):
        if self.null:
            return None
        return self.x2, self.y2

    @x1.setter
    def x1(self, value):
        self._set_geom_part(0, value)

    @y1.setter
    def y1(self, value):
        self._set_geom_part(1, value)

    @x2.setter
    def x2(self, value):
        self._set_geom_part(2, value)

    @y2.setter
    def y2(self, value):
        self._set_geom_part(3, value)

    @start.setter
    def start(self, value):
        if not isinstance(value, (list, tuple)) or len(value) != 2:
            raise OrmValueError('Vector.start must be set using list|tuple (x,y)!')
        self.x1, self.y1 = value

    @end.setter
    def end(self, value):
        if not isinstance(value, (list, tuple)) or len(value) != 2:
            raise OrmValueError('Vector.end must be set using list|tuple (x,y)!')
        self.x2, self.y2 = value

    def _compose_value_from_geometry(self):
        self._value = "(%sx%s),(%sx%s)" % (self._geometry[0], self._geometry[1], self._geometry[2], self._geometry[3])


class ValueGeoCoordinate(ValueGeometryAbstract):
    pass




# -*- coding: utf8 -*-
""" Different functions for general purpose. """

import re
import socket
import binascii
import datetime
import calendar


def gettext(v, context=None):
    """
    :v                              (str) Source text to translate to, in english;
    :context                        (str) Clarification about source text, if necessary;
    :returns                        (str) Translated text for given source english text and given
                                    context, if any gettext method is defined for this ORM
                                    configuration; or source text itself otherwise.
    """
    from .manage import manager
    f = manager.get_gettext_callback()
    return f(v, context) if callable(f) else v


def remove_leading_slash(path):
    """
    :path                           (str) Source path;
    :returns                        (str) Path without leading slash: if any is present - removing
                                    it, otherwise just returns source string.
    """
    if not path:
        return path
    if not isinstance(path, str):
        return path
    return path if path[0] != '/' else path[1:]


def ensure_leading_slash(path):
    """
    :path                           (str) Source path;
    :returns                        (str) Path with leading slash: if is not present - adding
                                    it, otherwise just returns source string.
    """
    if not path:
        return path
    if not isinstance(path, str):
        return path
    return path if path[0] == '/' else ("/%s" % path)


def remove_trailing_slash(path):
    """
    :path                           (str) Source path;
    :returns                        (str) Path without trailing slash: if any is present - removing
                                    it, otherwise just returns source string.
    """
    if not path:
        return path
    if not isinstance(path, str):
        return path
    return path if path[-1] != '/' else path[:-1]


def ensure_trailing_slash(path):
    """
    :path                           (str) Source path;
    :returns                        (str) Path with trailing slash: if is not present - adding
                                    it, otherwise just returns source string.
    """
    if not path:
        return path
    if not isinstance(path, str):
        return path
    return path if path[-1] == '/' else ("%s/" % path)


def path_slashed(path):
    return ensure_leading_slash(ensure_trailing_slash(path))


def path_leading_slash(path):
    return ensure_leading_slash(remove_trailing_slash(path))


def path_trailing_slash(path):
    return ensure_trailing_slash(remove_leading_slash(path))


def path_not_slashed(path):
    return remove_trailing_slash(remove_leading_slash(path))


def fit_image(image, **kwargs):
    """ Resizes given image to fit to given parameters: maximal or minimal
    width and/or height. If minimal and maximal values conflicts with
    each other on the given image - the maximal limitations will be taken.
    :image                  (image) The original image to resize to;
    :max_width              [int] The maximum width of resulting image;
    :max_height             [int] The maximum height of resulting image;
    :min_width              [int] The minimum width of resulting image;
    :min_height             [int] The minimum height of resulting image;
    :returns                (image) The new image object. Even if there is no
                            need to resize at all - the copy of original image will
                            be returned.
    """
    from PIL import Image

    max_width = kwargs.get('max_width', None)
    min_width = kwargs.get('min_width', None)
    max_height = kwargs.get('max_height', None)
    min_height = kwargs.get('min_height', None)
    img_width, img_height = image.size
    if not ((max_width and img_width > max_width)
            or (max_height and img_height > max_height)
            or (min_width and img_width < min_width)
            or (min_height and img_height < min_height)):
        return image.copy()
    mult_max_w = float(max_width) / float(img_width) if max_width and img_width > max_width else 1.0
    mult_max_h = float(max_height) / float(img_height) if max_height and img_height > max_height else 1.0
    mult_min_w = float(img_width) / float(min_width) if min_width and img_width < min_width else 1.0
    mult_min_h = float(img_height) / float(min_height) if min_height and img_height < min_height else 1.0
    mult_max = float(min(mult_max_w, mult_max_h))
    mult_min = float(max(mult_min_w, mult_min_h))
    mult = mult_max if mult_max != 1.0 else mult_min
    if mult == 1.0:
        return image.copy()
    r_width = int(float(img_width) * float(mult))
    r_height = int(float(img_height) * float(mult))
    return image.resize([r_width, r_height], Image.BICUBIC)


def image_thumbnail(image, width, height, crop=True, stretch=True):
    """ Prepares and returns a thumbnail of given image. The original
    image will not be affected.
    :image                  (image) The original image;
    :width                  (int) The width of resulting thumbnail (in pixels);
    :height                 (int) The height of resulting thumbnail (in pixels);
    :crop                   (bool) if set to 'True' - the image will be
                            resized that way to fill entire thumbnail box, even if
                            a part of it will let outside of borders. For example, for
                            portrait oriented image and boxed thumbnail, top and
                            bottom stripes will be cut off; if set to 'False' - the
                            image will be resized that way to fully fit in the given
                            sizes - width and height, and some space may left white
                            spaced;
                            (default=True)
    :stretch                (bool) if set to 'True' and the size of the
                            original image is smaller, than required size of thumbnail -
                            image will be zoomed. For example, if the original image has
                            size 30x30 pixels and thumbnail is about to be 64x64 pixels
                            size - the image will be zoomed (stretched) from 30x30 to
                            64x64. If set to 'False' and image is smaller than required
                            size of thumbnail - the size of image will not be changed.
                            (default=True)
    :returns                (image) The resulting thumbnail image object.
    """
    from PIL import Image

    img_width, img_height = image.size
    # If size of the original image is the same as required - just
    # return original image
    if img_width == width and img_height == height:
        return image.copy()
    # If size (both width and height) is smaller than required and
    # 'stretch' is not set_value to 'True' - return original image
    if img_width <= width and img_height <= height and not stretch:
        return image.copy()
    # Resizing image to the required sizes
    mult = 1.0
    # If stretch is required
    if (img_width < width or img_height < height) and stretch:
        mult_w = float(width) / float(img_width)
        mult_h = float(height) / float(img_height)
        mult = float(max(mult_w, mult_h))
    r_img_width = int(float(img_width) * float(mult))
    r_img_height = int(float(img_height) * float(mult))
    # If resize is required. Depending on the existing multiplier, which is
    # 1.0 by default (no strech affected) and may be modified by stretch mechanics
    if r_img_width > width or r_img_height > height:
        mult_w = float(width) / float(r_img_width)
        mult_h = float(height) / float(r_img_height)
        mult *= float(min((max(mult_w, mult_h) if crop else min(mult_w, mult_h)), 1.0))
        r_img_width = int(float(img_width) * float(mult))
        r_img_height = int(float(img_height) * float(mult))
    thumbnail = image.resize([r_img_width, r_img_height], Image.BICUBIC)
    img_width, img_height = thumbnail.size
    # If crop is required and the picture is not boxed - do crop
    if crop and (img_width != img_height):
        if img_width > img_height:
            r_top = 0
            r_bottom = img_height - 1
            r_left = (img_width / 2) - (width / 2)
            r_right = (img_width / 2) + (width / 2)
        else:
            r_left = 0
            r_right = 0
            r_top = (img_height / 2) - (height / 2)
            r_bottom = (img_height / 2) + (height / 2)
        thumbnail = thumbnail.crop([r_left, r_top, r_right, r_bottom])
    return thumbnail


def split_search_phrase(phrase):
    """ Splits the search phrase (one solid string of text) into the search words. Words can be
    simple - """
    if not isinstance(phrase, str):
        raise TypeError("'split_findphrase' requires that parameter to be a (str)!")
    if not phrase:
        return []
    r = []
    vv = re.findall(r'[^"\s]\S*|".+?"', phrase.replace('\'', '"').replace('""', '"'))
    for v in vv:
        v_ = re.sub("[!\(\)\-\_\+\=\[\]\{\}\/\"\\\:\;\,\.\?\<\>]", "", v)
        if not v_:
            continue
        r.append(str(v).replace('"', ''))
    return r


def ipaddr_version(val):
    """ Checks the version of IP address, returning numeric 4 for IPv4, numeric 6 for IPv6 or 0 otherwise.
    :val                            (str) Source string containing the IP address;
    :returns                        (int) returns '4' for IPv4, '6' for IPv6, '0' otherwise.
    """
    if not isinstance(val, str):
        raise TypeError("'ipaddr_version' requires value to be type of (str)!")
    if '.' in val and val.count('.') == 3:
        p = val.split('.')
        if not p[0].isdigit() or not p[1].isdigit() or not p[2].isdigit() or not p[3].isdigit():
            return 0
        if int(p[0]) > 255 or int(p[1]) > 255 or int(p[2]) > 255 or int(p[3]) > 255:
            return 0
        return 4
    elif ':' in val and val.count('::') <= 1:
        p = val.split(':')
        if len(p) > 8:
            return 0
        for p_ in p:
            if p_ == "":
                continue
            if len(p_) > 4:
                return 0
            if re.sub("[^0-9a-fA-F]", "", str(p_)) != p_:
                return 0
        return 6
    else:
        return 0


def hex2ip(val):
    """ Converts heximal IP address to the corresponding textual format.
    :val                            (str) Source heximal IP address record;
    :returns                        (str) Textual IP address,
                                    (None) if incorrect value was given.
    """
    try:
        val = str(val)
        if len(val) == 8:
            return socket.inet_ntop(socket.AF_INET, binascii.unhexlify(val))
        elif len(val) == 32:
            return socket.inet_ntop(socket.AF_INET6, binascii.unhexlify(val))
        else:
            return None
    except (binascii.Error, OSError):
        return None


def ip2hex(val):
    """ Converts textual IP address (IPv4 or IPv6) to the heximal string.
    :returns                        (str) Heximal record of the given IP address;
                                    (None) if invalid address or empty value given.
    """
    if not val:
        return None
    ipvers = ipaddr_version(val)
    try:
        if ipvers == 0:
            return None
        elif ipvers == 4:
            return binascii.hexlify(socket.inet_pton(socket.AF_INET, val)).upper()
        elif ipvers == 6:
            return binascii.hexlify(socket.inet_pton(socket.AF_INET6, val)).upper()
        else:
            return None
    except (binascii.Error, OSError):
        return None


def int_2_ipv4(val):
    """ Converts integer value to the textual IPv4 address. Not working with IPv6, use
    heximal values instead.
    :val                            (int) Source integer reprsents IP address;
    :returns                        (str) textual IPv4 record.
    """
    val = int(val)
    o1 = int(val / 16777216) % 256
    o2 = int(val / 65536) % 256
    o3 = int(val / 256) % 256
    o4 = val % 256
    return "%i.%i.%i.%i" % (o1, o2, o3, o4)


def ipv4_2_int(val):
    """ Converts textual IPv4 to integer. Not working with IPv6, use heximal instead.
    :val                            (str) Source textual IPv4 record;
    :returns                        (int) Integer representation of IPv4 octets;
                                    (None) if failed.
    """
    if ipaddr_version(val) != 4:
        return None
    return int(ip2hex(val), 16)


def int2mac(val, separator='.'):
    """ Converts integer to textual network L2 MAC address.
    :val                            (int) Source long integer represents MAC address;
    :separator                      (str) Separator which will be used to divide MAC
                                    words between themselfs;
                                    (default='.')
    :returns                        (str) String representation of MAC address.
    """
    mac_ = "%0.12x" % val
    if not separator:
        return mac_
    elif separator == '.':
        return "%s.%s.%s" % (str(mac_[0:4]), str(mac_[4:8]), str(mac_[8:12]))
    else:
        return separator.join([
            mac_[0:2],
            mac_[2:4],
            mac_[4:6],
            mac_[6:8],
            mac_[8:10],
            mac_[10:12]
        ])


def ipv4mask_2_cidr(netmask):
    """ Converts a.b.c.d netmask to the /CIDR format.
    For example: 255.255.255.0 will be converted to the /24 (returned as integer 24).
    :netmask                        (str) Source netmask in A.B.C.D format;
    :returns                        (int) CIDR netmask.
    """
    return sum([bin(int(x)).count('1') for x in netmask.split('.')])


def format_datetime(value, format_):
    """
    Converts Python datetime value to the textual one, using given format. Using 'gettext'
    functionality to try to translate english naming to the localized one.

    Format legend:
        %d      day of month with leading zero [01..31]
        %j      day of month without leading zero [1..31]
        %D      3-symbols day of week [Mon..Sun]
        %l      fully-named day of week [Monday..Sunday]
        %N      number of day of week in ISO format [1..7]
        %z      serial number of day of year, starting from zero [0..365]
        %W      serial number of week of year, starting from one [1..]
        %M      3-symbols name of month [Jan..Dec]
        %F      fully named month, [January..December]
        %m      month with leading zero [01..12]
        %n      month without leading zero [1..12]
        %t      quantity of days in month [..31]
        %Y      4-digits year
        %y      2-digits year
        %a      am/pm
        %A      AM/PM
        %g      hours in 12-hours format without leading zero
        %G      hours in 24-hours format without leading zero
        %h      hours in 12-hours format with leading zero
        %H      hours in 24-hours format with leading zero
        %i      minutes with leading zero [00..59]
        %s      seconds with leading zero [00..59]

    :value                          (datetime) Source value to be converted to the string
                                    using given format;
    :format_                        (str) String containing format which must be used to
                                    format given source date/time value;
    :returns                        (str) String with date/time formatted by given format.
    """
    if not isinstance(format_, str):
        raise TypeError("'format' must be a (str) type!")
    elif not format_:
        raise ValueError("'format' must be set and not be an empty string!")
    v_day = getattr(value, 'day', None)
    v_month = getattr(value, 'month', None)
    v_year = getattr(value, 'year', None)
    v_hours = getattr(value, 'hour', None)
    v_minutes = getattr(value, 'minute', None)
    v_seconds = getattr(value, 'second', None)
    v_weekday = value.isoweekday() if hasattr(value, 'isoweekday') else None
    b_date = bool(v_day is not None and v_month is not None and v_year is not None)
    b_time = bool(v_hours is not None and v_minutes is not None and v_seconds is not None)

    f = dict()
    # %d :: 2-digits day [01..31]
    f['d'] = "%02i" % int(v_day) if b_date else "00"
    # %D :: 3-symbols day of week representation [Mon..Sun]
    if not b_date:
        f['D'] = ""
    else:
        t_weekdays3 = [
            gettext('Mon'),
            gettext('Tue'),
            gettext('Wed'),
            gettext('Thu'),
            gettext('Fri'),
            gettext('Sat'),
            gettext('Sun')
        ]
        f['D'] = t_weekdays3[v_weekday - 1]
    # %j :: day without leading zero [1..31]
    f['j'] = str(v_day) if b_date else "0"
    # %l :: fully named way of week (lowercase L) [Monday..Sunday]
    if not b_date:
        f['l'] = ""
    else:
        t_weekdays = [
            gettext('Monday'),
            gettext('Tuesday'),
            gettext('Wednesday'),
            gettext('Thursday'),
            gettext('Friday'),
            gettext('Saturday'),
            gettext('Sunday')
        ]
        f['l'] = t_weekdays[v_weekday-1]
    # %N :: number of day of week in ISO format [1..7]
    f['N'] = str(v_weekday-1) if b_date else ""
    # %z :: serial number of day of year, starting from zero [0..365]
    if not b_date:
        f['z'] = "0"
    elif isinstance(value, datetime.date) and not isinstance(value, datetime.datetime):
        f['z'] = str((value - datetime.date(value.year, 1, 1)).days)
    else:
        f['z'] = str((value - datetime.datetime(value.year, 1, 1)).days)
    # %W :: serial number of week of year, starting from one [1..]
    f['W'] = "0" if not b_date else str(value.isocalendar()[1])
    # %F :: fully named month, [January..December]
    if not b_date:
        f['F'] = ""
    else:
        t_months = [
            gettext('January'),
            gettext('February'),
            gettext('March'),
            gettext('April'),
            gettext('May'),
            gettext('June'),
            gettext('July'),
            gettext('August'),
            gettext('September'),
            gettext('October'),
            gettext('November'),
            gettext('December')
        ]
        f['F'] = t_months[v_month - 1]
    # %m :: Month with leading zero [01..12]
    f['m'] = "%02i" % int(v_month) if b_date else "00"
    # %M :: 3-symbols name of month [Jan..Dec]
    if not b_date:
        f['M'] = ""
    else:
        t_months3 = [
            gettext('Jan'),
            gettext('Feb'),
            gettext('Mar'),
            gettext('Apr'),
            gettext('May'),
            gettext('Jun'),
            gettext('Jul'),
            gettext('Aug'),
            gettext('Sep'),
            gettext('Oct'),
            gettext('Nov'),
            gettext('Dec')
        ]
        f['M'] = t_months3[v_month - 1]
    # %n :: Month without leading zero [1..12]
    f['n'] = str(v_month) if b_date else "0"
    # %t :: Quantity of days in month [..31]
    f['t'] = str(calendar.monthrange(v_year, v_month)) if b_date else '0'
    # %Y :: 4-digits year
    f['Y'] = str(v_year) if b_date else "0000"
    # %y :: 2-digits year
    f['y'] = str(v_year)[2:] if b_date else "00"
    # %a :: am/pm
    f['a'] = "" if not b_time else ("am" if v_hours < 12 else "pm")
    # %A :: AM/PM
    f['A'] = "" if not b_time else ("AM" if v_hours < 12 else "PM")
    # %g :: hours in 12-hours format without leading zero
    f['g'] = "0" if not b_time else str(12 if v_hours == 0 else (v_hours if v_hours < 12 else (v_hours - 12)))
    # %G :: hours in 24-hours format without leading zero
    f['G'] = "0" if not b_time else str(v_hours)
    # %h :: hours in 12-hours format with leading zero
    f['h'] = "00" if not b_time else "%02i" % int(12 if v_hours == 0 else (v_hours if v_hours < 12 else (v_hours - 12)))
    # %H :: hours in 24-hours format with leading zero
    f['H'] = "00" if not b_time else "%02i" % int(v_hours)
    # %i :: minutes with leading zero [00..59]
    f['i'] = "00" if not b_time else "%02i" % int(v_minutes)
    # %s :: seconds with leading zero [00..59]
    f['s'] = "00" if not b_time else "%02i" % int(v_seconds)

    formatted = format_
    for k in f:
        v = f[k]
        if v is None:
            continue
        k_ = "%" + str(k)
        formatted = formatted.replace(k_, v)

    return formatted


def _str2datetime(value, hasdate, hastime):
    """
    Converts given string, containing date, time or both date and time, to the corresponding
    Python date, time or datetime typed value.
    :value                          (str) Source string;
    :hasdate                        (bool) Has source string date or not;
    :hastime                        (bool) Has source string time or not;
    :returns                        (datetime) if given source has both date and time;
                                    (date) if given source has date only;
                                    (time) if given source has time only;
    """
    ps = value.split(' ')
    d_ = False
    t_ = False
    n_year = 0
    n_month = 0
    n_day = 0
    n_hour = 0
    n_minute = 0
    n_second = 0
    for p in ps:
        if not p:
            continue
        if '.' not in p and '/' not in p and '\\' not in p and '-' not in p and ':' not in p:
            raise ValueError("Invalid (str) datetime format!")
        if ':' in p:
            if t_:
                raise ValueError("Invalid (str) datetime format!")
            if p.count(':') > 2:
                raise ValueError("Invalid (str) datetime format!")
            p_ = p.split(':')
            if not p_[0].isdigit() or not p_[1].isdigit() or (len(p_) == 3 and not p_[2].isdigit()):
                raise ValueError("Invalid (str) datetime format!")
            n_hour = int(p_[0])
            n_minute = int(p_[1])
            n_second = int(p_[2]) if len(p) == 3 else 0
            if n_hour > 23 or n_minute > 59 or n_second > 59:
                raise ValueError("Invalid (str) datetime format!")
            t_ = True
        else:
            if d_:
                raise ValueError("Invalid (str) datetime format!")
            s = '.' if '.' in p else ('-' if '-' in p else ('/' if '/' in p else '\\'))
            p_ = p.split(s)
            if len(p_) != 3:
                raise ValueError("Invalid (str) datetime format!")
            if not p_[0].isdigit() or not p_[1].isdigit() or not p_[2].isdigit():
                raise ValueError("Invalid (str) datetime format!")
            p1 = int(p_[0])
            p2 = int(p_[1])
            p3 = int(p_[2])
            starting_with_year = p1 > 31 or s == '-'
            if starting_with_year:
                n_year = p1
                if s in ('.', '-'):
                    n_month = p2
                    n_day = p3
                else:
                    n_month = p3
                    n_day = p2
            else:
                n_year = p3
                if s in ('.', '-'):
                    n_month = p2
                    n_day = p1
                else:
                    n_month = p1
                    n_day = p2
            if 99 < n_year < 1000:
                raise ValueError("Invalid (str) datetime format!")
            if n_year < 1000:
                n_year = n_year + 1000 if n_year > 69 else n_year + 2000
            if n_day == 0 or n_month == 0 or n_day > 31 or n_month > 12:
                raise ValueError("Invalid (str) datetime format!")
            n_daysmax = calendar.monthrange(n_year, n_month)[1]
            if n_day > n_daysmax:
                raise ValueError("Invalid (str) datetime format!")
            d_ = True
    if hasdate and not hastime:
        if not d_:
            raise ValueError("Invalid (str) datetime format!")
        value = datetime.date(n_year, n_month, n_day)
    elif not hasdate and hastime:
        if not t_:
            raise ValueError("Invalid (str) datetime format!")
        value = datetime.time(n_hour, n_minute, n_second)
    else:
        if not d_:
            raise ValueError("Invalid (str) datetime format!")
        value = datetime.datetime(n_year, n_month, n_day, n_hour, n_minute, n_second)
    return value


def str2datetime(value):
    """ Converts given string, containing date and time, to Python datetime type.
    :value                          (str) Source textual date-time;
    :returns                        (time) Python datetime-typed value.
    """
    return _str2datetime(value, True, True)


def str2date(value):
    """ Converts given string, containing date, to Python date type.
    :value                          (str) Source textual date;
    :returns                        (time) Python date-typed value.
    """
    return _str2datetime(value, True, False)


def str2time(value):
    """ Converts given string, containing time, to Python time type.
    :value                          (str) Source textual time;
    :returns                        (time) Python time-typed value.
    """
    return _str2datetime(value, False, True)




import datetime
from ..exceptions import OrmNotDefinedYet
from ..utils import str2datetime, str2date, str2time


def wrap_value(model, instance, c, key, value, textual=False):
    from ..funcs import RequestTableBelongedColumn
    if type(value) is AttributeValueNull:
        value = None
    if isinstance(c, RequestTableBelongedColumn):
        c = c.c
    value_class = getattr(c, 'value_class', None)
    value_handler = getattr(c, 'value_handler', None)
    if value_handler is not None:
        value = value_handler(value)
    if value_class is not None and not (textual and isinstance(value, str)):
        if AttributeValueAbstract in value_class.mro():
            if value_class.basecls is not False:
                if value is None and not c.nullable:
                    value = c.default
                value = value_class.new(value) if value is not None else AttributeValueNull()
            else:
                value = value_class.new(value)
        else:
            value = value_class(value)
        value.model = model
        value.model_k = key
        value.instance = instance
        value.c = c
    elif isinstance(value,
                    (str, int, float, bytes, bool, tuple, list, datetime.date, datetime.time, datetime.datetime)) \
            or value is None:
        if isinstance(value, str):
            value_class = AttributeValueString
        elif isinstance(value, int):
            value_class = AttributeValueInteger
        elif isinstance(value, float):
            value_class = AttributeValueFloat
        elif isinstance(value, bytes):
            value_class = AttributeValueBinary
        elif isinstance(value, bool):
            value_class = AttributeValueBoolean
        elif isinstance(value, datetime.datetime) \
                and not isinstance(value, datetime.date) \
                and not isinstance(value, datetime.time):
            value_class = AttributeValueDateTime
        elif isinstance(value, datetime.date):
            value_class = AttributeValueDate
        elif isinstance(value, datetime.time):
            value_class = AttributeValueTime
        else:
            value_class = AttributeValueNull
        value = value_class.new(value)
        value.model = model
        value.model_k = key
        value.instance = instance
        value.c = c
    return value


class AttributeValueAbstract(object):
    """ Abstract class for attribute value representation types """
    basecls = None
    check_modified = True

    def __init__(self, *args, **kwargs):
        self.model = None
        self.model_k = None
        self.instance = None
        self.c = None

    @classmethod
    def new(cls, value):
        """ Calls by ORM mechanics right after creating the class'es instance. """
        return cls(value)

    def initialize(self, key, value):
        self.c.initialize(self.instance, key, value)

    @property
    def caption(self):
        return self.c.caption

    @property
    def abbreviation(self):
        return self.c.abbreviation

    @property
    def options(self):
        return self.c.options

    @property
    def string(self):
        if not hasattr(self, 'text_repr'):
            return str(self.base(self))
        return self.text_repr(self.base(self))

    def base(self, value):
        if self.basecls is None or isinstance(self.basecls, bool):
            return None
        return self.basecls(value)

    @property
    def value(self):
        return self.base(self)

    def __repr__(self):
        r = self.string
        return str(r) if not isinstance(r, str) else r

    def text_repr(self, value):
        if not hasattr(self.c, 'text_repr'):
            return str(value)
        r = self.c.text_repr(value, instance=self.instance)
        return "" if r is None else str(r)

    def log_repr(self, src, dst):
        return self.c.log_repr(src, dst, instance=self.instance)

    def call_on_change(self, src, dst):
        self.c.call_on_change(src, dst)

    def verify(self, value):
        return self.c.verify(value)


class AttributeValueNull(AttributeValueAbstract):
    """ Representation of 'NULL' value ('None' in Python terms) """
    basecls = None

    @classmethod
    def new(cls, value):
        return cls()

    @property
    def string(self):
        return self.text_repr(None)

    def base(self, value):
        return None

    @property
    def value(self):
        return None

    def __repr__(self):
        return "<NULL>"

    def __nonzero__(self):
        return False

    def __bool__(self):
        return False

    def log_repr(self, src, dst):
        return self.c.log_repr(src, dst, instance=self.instance)

    def on_change(self, src, dst):
        self.c.on_change(src, dst)

    def verify(self, value):
        return self.c.verify(value)


class AttributeValueString(AttributeValueAbstract, str):
    basecls = str


class AttributeValueBinary(AttributeValueAbstract, bytes):
    basecls = bytes


class AttributeValueBoolean(AttributeValueAbstract, int):
    basecls = int

    @property
    def value(self):
        return bool(self.base(self))

    def __str__(self):
        return "True" if self.value else "False"


class AttributeValueInteger(AttributeValueAbstract, int):
    basecls = int


class AttributeValueFloat(AttributeValueAbstract, float):
    basecls = float


class AttributeValueDate(AttributeValueAbstract, datetime.date):
    basecls = datetime.date

    def base(self, value):
        if value is None or isinstance(value, AttributeValueNull):
            return None
        elif isinstance(value, (datetime.datetime, datetime.date)):
            return datetime.date(value.year, value.month, value.day)
        elif isinstance(value, str):
            try:
                v_ = str2date(value)
            except ValueError:
                return None
            return datetime.date(v_.year, v_.month, v_.day)
        return None

    @classmethod
    def new(cls, value):
        if value is None:
            return AttributeValueNull()
        if isinstance(value, (datetime.datetime, datetime.date)):
            return cls(value.year, value.month, value.day)
        elif isinstance(value, str):
            try:
                v_ = str2date(value)
            except ValueError:
                return AttributeValueNull()
            return cls(v_.year, v_.month, v_.day)
        return AttributeValueNull()


class AttributeValueTime(AttributeValueAbstract, datetime.time):
    basecls = datetime.time

    def base(self, value):
        if value is None or isinstance(value, AttributeValueNull):
            return None
        elif isinstance(value, (datetime.datetime, datetime.time)):
            return datetime.time(value.hour, value.minute, value.second)
        elif isinstance(value, str):
            try:
                v_ = str2time(value)
            except ValueError:
                return None
            return datetime.time(v_.hour, v_.minute, v_.second)
        return None

    @classmethod
    def new(cls, value):
        if value is None:
            return AttributeValueNull()
        if isinstance(value, (datetime.datetime, datetime.time)):
            return cls(value.hour, value.minute, value.second)
        elif isinstance(value, str):
            try:
                v_ = str2time(value)
            except ValueError:
                return AttributeValueNull()
            return cls(v_.hour, v_.minute, v_.second)
        return AttributeValueNull()


class AttributeValueDateTime(AttributeValueAbstract, datetime.datetime):
    basecls = datetime.datetime

    def base(self, value):
        if value is None or isinstance(value, AttributeValueNull):
            return None
        elif isinstance(value, datetime.date) and not isinstance(value, datetime.datetime):
            return datetime.datetime(value.year, value.month, value.day, 0, 0, 0)
        elif isinstance(value, datetime.datetime):
            return datetime.datetime(value.year, value.month, value.day, value.hour, value.minute, value.second)
        elif isinstance(value, str):
            try:
                v_ = str2datetime(value)
            except ValueError:
                return None
            return datetime.datetime(v_.year, v_.month, v_.day, v_.hour, v_.minute, v_.second)
        return None

    @classmethod
    def new(cls, value):
        if value is None:
            return AttributeValueNull()
        if isinstance(value, datetime.date) and not isinstance(value, datetime.datetime):
            return cls(value.year, value.month, value.day, 0, 0, 0)
        elif isinstance(value, datetime.datetime):
            return cls(value.year, value.month, value.day, value.hour, value.minute, value.second)
        elif isinstance(value, str):
            try:
                v_ = str2datetime(value)
            except ValueError:
                return AttributeValueNull()
            return cls(v_.year, v_.month, v_.day, v_.hour, v_.minute, v_.second)
        return AttributeValueNull()


class AttributeValueVirtual(AttributeValueAbstract):
    check_modified = False

    def __init__(self):
        super(AttributeValueVirtual, self).__init__()
        self.instance = None
        self.model_k = None
        self.c = None
        self._value = None
        self._loaded = False

    def get_value(self):
        if self._loaded is False:
            self._value = self.load()
        return self._value

    def set_value(self, instance, key, value):
        return value

    def load(self):
        pass

    def reload(self):
        self._value = self.load()
        return self._value

    def reset(self):
        self._loaded = False
        self._value = None

    def pretend_loaded(self):
        self._loaded = True

    @property
    def loaded(self):
        return self._loaded


class ValueArrayAbstract(AttributeValueVirtual):
    basecls = False
    check_modified = False

    def __init__(self):
        super(ValueArrayAbstract, self).__init__()
        self._original_values = None
        self._values = None
        self._objects = None
        self._to_delete = list()
        self.reset()

    def __repr__(self):
        return "Array"

    def _get_c(self):
        return getattr(self.instance.__class__, self.model_k) if self.model_k is not None else None

    def _verify_changes(self):
        if self._original_values == self._values and self.model_k in self.instance.__changes__:
            del(self.instance.__changes__[self.model_k])
        elif self._original_values != self._values:
            self.instance.__changes__[self.model_k] = (self._original_values, self._values)

    def write(self):
        for obj in self._to_delete:
            obj.delete()
        if isinstance(self._objects, dict):
            for k in self._objects:
                obj = self._objects[k]
                obj.write(session=self.instance.__session__)
        else:
            for obj in self._objects:
                obj.write(session=self.instance.__session__)
        self._to_delete = list()

    def load(self):
        pass

    def ensure(self):
        if self._loaded:
            return
        self.load()

    def reload(self):
        self.reset()
        self.load()

    def reset(self):
        pass

    def get_value(self):
        self.ensure()
        return self

    def set_value(self, instance, key, value):
        pass

    def _assign_with_instance(self, obj):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot access values of virtual related iterable until model been defined!")
        model = c.model
        if model is None:
            raise OrmNotDefinedYet("cannot access values of virtual related iterable until model been defined!")
        pk = model.__meta__.primary_key
        for k in pk:
            setattr(obj, "parent_%s" % k, getattr(self.instance, k))





from setuptools import setup, find_packages


setup(
    name='genorm',
    version='0.1.7',
    description='GenORM (simple functional MySQL/PostgreSQL ORM)',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Database :: Front-Ends',
    ],
    keywords='orm mysql psql postgres postgresql genorm',
    url='https://bitbucket.org/deniskhodus/genorm',
    author='Denis Khodus',
    author_email='deniskhodus@gmail.com',
    license='MIT',
    packages=find_packages(),
    install_requires=['Pillow', 'pc23'],
    include_package_data=False,
    zip_safe=False,
)




# -*- coding: utf8 -*-
""" Custom ORM defined exception classes. """


class OrmException(Exception):
    """ Base class for ORM exceptions. """
    pass


class OrmManagementError(OrmException):
    pass


class OrmDefinitionError(OrmException):
    pass


class OrmSessionNotSet(OrmException):
    """ Raises when ORM does not knows which session to use for the object. Programmer must specify
    the session using kwargs argument 'session=' or set default session using manager's method
    'set_session(<session>)'.
    """
    pass


class OrmModelClassMismatch(OrmException):
    """ Mismatch between offered via argument Model and expected one. """
    pass


class OrmAttributeTypeMistake(OrmException):
    """ Mismatch between used by programmer attribute type and expected one. """
    pass


class OrmParameterError(OrmException):
    """ All various parameter errors (type mismatchs, not enough arguments and so on)
    raises this type of exception. """
    pass


class OrmInvalidKey(OrmException):
    pass


class OrmModelDeclarationError(OrmException):
    pass


class OrmRequestError(OrmException):
    pass


class OrmInstanceMustBeWritten(OrmException):
    pass


class OrmInstanceNotFound(OrmException):
    pass


class OrmAttributeIsProtected(OrmException):
    pass


class OrmComparatorValueError(OrmException):
    pass


class OrmValueError(OrmException):
    pass


class OrmTypeError(OrmException):
    pass


class OrmUnacceptableOperation(OrmException):
    """ Raises when trying to do something with the ORM element, which is normal for
    the base type (for example, for list or dict), but is not acceptable for that type's
    ORM custom implementation (for example, for files and pictures lists). """
    pass


class OrmNotDefinedYet(OrmException):
    """ Raises when trying to deal with ORM elements such as attributes while according
    Model definition not finished yet. """
    pass


class OrmLoggingError(OrmException):
    """ Various errors caused in the journaling logics. """
    pass


class OrmRowVersionMismatch(OrmException):
    """ Raises when submitting object been previously changed by other
    process in a time between this object been loaded from the database and
    object updating on the DBMS side, to inform about split brain situation
    avoiding (soft lock mechanism). """
    pass


class OrmProhibittedValueSet(OrmException):
    """ Internal exception, used in the model and engine mechanics to
    inform that mechanics that given value cannot be assigned to. """
    pass


class OrmPasswordIsWeak(OrmException):
    """ Raises when given password is too weak, for (Password)
    attribute class. """
    pass


class OrmAttachmentAlreadyExists(OrmException):
    """ Raises when attachment cannot be saved to the according storage
    due another file with the same name is already exists and attaching
    file cannot be renamed. """
    pass


class OrmAttachmentSourceNotExists(OrmException):
    """ Raises when source file, given in 'path' parameter, is not
    exists at the file system (trying to attach from not existing
    file). """
    pass


class OrmAttachmentNotExists(OrmException):
    """ Raises when the trying to act with attachment - for example,
    save it to the another location of file system, - but attachment
    is not exists (not attached to the attribute). """
    pass


class OrmObjectMustBeWrittenFirst(OrmException):
    """ Raises when trying to deal with object instance which did not
    been written to the database (not INSERTed) before. Some methods,
    data types and functions can only work with already existing
    in the database objects (for example - they can relly on already
    existing ID of object). """
    pass


class OrmValueVerificationError(OrmException):
    pass




import os
import os.path
import shutil
import math
from .base import AttributeValueVirtual, AttributeValueNull
from ..exceptions import OrmValueError, OrmNotDefinedYet, OrmParameterError, \
    OrmAttachmentAlreadyExists, OrmAttachmentNotExists
from ..utils import gettext


class ValueAttachmentAbstract(AttributeValueVirtual):
    basecls = False
    check_modified = True

    def __init__(self, value):
        super(ValueAttachmentAbstract, self).__init__()
        self._attachment = None
        self._on_write = False
        self._value = value
        self._need_write = False
        self.parent_instance = None
        self.parent_k = None
        self.attributes_instance = None

    def __repr__(self):
        return "<Not attached>" if not self.is_attached else "Attachment(%s)" % self._value

    def base(self, value):
        return None if value is None or isinstance(value, AttributeValueNull) else str(value)

    @property
    def value(self):
        return self._value

    @property
    def string(self):
        return str(self._value) if self.is_attached else ""

    def _get_c(self):
        return getattr(self.instance.__class__, self.model_k) if self.model_k is not None else None

    def get_value(self):
        return self

    @property
    def _instance(self):
        return self.instance if not self.parent_instance else self.parent_instance

    @property
    def _attr_name(self):
        return self.model_k if not self.parent_k else self.parent_k

    @property
    def is_attached(self):
        return self._value is not None

    def set_value(self, instance, key, value):
        before_ = self._value
        if isinstance(value, ValueAttachmentAbstract):
            self._value = value.value
            return self._value, before_
        if value is None or isinstance(value, AttributeValueNull):
            self._value = None
            return None, before_
        if not isinstance(value, str):
            raise OrmValueError("single attachments can be directly set specifying (str) file name in the storage!")
        self._value = value
        self._attachment = self.abs_path
        return self._value, before_

    def verify_storage_existance(self):
        if self._value is None:
            return True
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        r = c.storage.verify_file_existance(self._instance, self._attr_name, self._value)
        if not r:
            self._value = None
        return r

    def _check_not_attached(self):
        if self.is_attached:
            raise OrmAttachmentAlreadyExists(self._attr_name)

    def _check_attached(self):
        if not self.is_attached:
            raise OrmAttachmentNotExists(self._attr_name)

    def _execute_on_write(self):
        if self._on_write is False:
            return
        if not isinstance(self._on_write, (list, tuple)):
            return
        f = self._on_write[0]
        args = self._on_write[1] if len(self._on_write) >= 2 else list()
        kwargs = self._on_write[2] if len(self._on_write) >= 3 else dict()
        self._on_write = False
        r = f(*args, **kwargs)
        if r is None:
            return
        self._value = r if not isinstance(r, (list, tuple)) else r[0]

    def before_instance_write(self):
        if not self._need_write:
            return
        if not self._instance.__existing__:
            return
        self._execute_on_write()
        self._need_write = False
        self._attachment = None

    def after_instance_write(self):
        if not self._need_write:
            return
        self._execute_on_write()
        self._need_write = False
        self._attachment = None

    def after_instance_deleted(self):
        if not self.is_attached:
            return
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        c.storage.del_file(self._instance, self._attr_name, str(self._value))

    def on_instance_load(self):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        if c.check_existance:
            self.verify_storage_existance()

    def push_file(self, src, filename=None, rule=None):
        from ..attributes.virtual import ATTACH_BY_COPY, ATTACH_BY_MOVE, ATTACH_BY_LINK
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        # self._check_not_attached()
        rule = rule or c.push_rule
        if rule not in (ATTACH_BY_COPY, ATTACH_BY_LINK, ATTACH_BY_MOVE):
            raise OrmParameterError(
                "attachment rule can be one of ATTACH_BY_COPY, ATTACH_BY_LINK or ATTACH_BY_MOVE constant!"
            )
        if rule == ATTACH_BY_COPY:
            self._copy_file_from(src, filename)
        elif rule == ATTACH_BY_LINK:
            self._link_file_from(src, filename)
        else:
            self._move_file_from(src, filename)

    def remove_file(self):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._check_attached()
        self._on_write = c.storage.del_file, (self._instance, self._attr_name, str(self._value))
        self._value = None
        self._attachment = None
        self._need_write = True
        self.instance.__changes__[self.model_k] = (gettext('Detached', 'attachments'), None)
        if self._instance != self.instance:
            self._instance.__changes__[self._attr_name] = (gettext('Detached', 'attachments'), None)

    def pull_file(self, dst, rule=None):
        from ..attributes.virtual import ATTACH_BY_COPY, ATTACH_BY_MOVE, ATTACH_BY_LINK
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        rule = rule or c.pull_rule
        if rule not in (ATTACH_BY_COPY, ATTACH_BY_LINK, ATTACH_BY_MOVE):
            raise OrmParameterError(
                "attachment rule can be one of ATTACH_BY_COPY, ATTACH_BY_LINK or ATTACH_BY_MOVE constant!"
            )
        if rule == ATTACH_BY_COPY:
            self._copy_file_to(dst)
        elif rule == ATTACH_BY_LINK:
            self._link_file_to(dst)
        else:
            self._move_file_to(dst)

    def _push_on_write_file(self, f, src, filename):
        fn, fp = f(src, self._instance, self._attr_name, filename)
        self._value = fn
        self.instance.__changes__[self.model_k] = ("%s '%s'" % (gettext('Attached', 'attachments'), fn), None)
        if self._instance != self.instance:
            self._instance.__changes__[self._attr_name] = ("%s '%s'" % (gettext('Attached', 'attachments'), fn), None)
        if self.attributes_instance is not None:
            setattr(self.attributes_instance, 'filesize', int(math.ceil(float(os.stat(fp).st_size) / 1024.0)))

    def _copy_file_from(self, src, filename=None):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._value = c.storage.suggest_filename(self._instance, self._attr_name, filename or src)
        self._on_write = self._push_on_write_file, (c.storage.copy_file_from, src, str(self._value))
        self._need_write = True
        self._attachment = src
        self.instance.__changes__[self.model_k] = (gettext('Attached', 'attachments'), None)
        if self._instance != self.instance:
            self._instance.__changes__[self._attr_name] = (gettext('Attached', 'attachments'), None)

    def _link_file_from(self, src, filename=None):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._value = c.storage.suggest_filename(self.instance, self.model_k, filename or src)
        self._on_write = self._push_on_write_file, (c.storage.link_file_from, src, str(self._value))
        self._need_write = True
        self._attachment = src
        self.instance.__changes__[self.model_k] = (gettext('Attached', 'attachments'), None)
        if self._instance != self.instance:
            self._instance.__changes__[self._attr_name] = (gettext('Attached', 'attachments'), None)

    def _move_file_from(self, src, filename=None):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._value = c.storage.suggest_filename(self.instance, self.model_k, filename or src)
        self._on_write = self._push_on_write_file, (c.storage.move_file_from, src, str(self._value))
        self._need_write = True
        self._attachment = src
        self.instance.__changes__[self.model_k] = (gettext('Attached', 'attachments'), None)
        if self._instance != self.instance:
            self._instance.__changes__[self._attr_name] = (gettext('Attached', 'attachments'), None)

    def _move_file_to(self, dst):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._check_attached()
        if dst[-1] == '/':
            dst += self._value
        if not self._need_write:
            c.storage.link_file_to(dst, self._instance, self._attr_name)
        else:
            if not os.path.lexists(self._attachment):
                raise OrmAttachmentNotExists(self._attachment)
            os.link(self._attachment, dst)
        self.remove_file()

    def _copy_file_to(self, dst):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._check_attached()
        if dst[-1] == '/':
            dst += self._value
        if not self._need_write:
            c.storage.copy_file_to(dst, self._instance, self._attr_name)
        else:
            if not os.path.lexists(self._attachment):
                raise OrmAttachmentNotExists(self._attachment)
            shutil.copyfile(self._attachment, dst)

    def _link_file_to(self, dst):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._check_attached()
        if dst[-1] == '/':
            dst += self._value
        if not self._need_write:
            c.storage.link_file_to(dst, self._instance, self._attr_name)
        else:
            if not os.path.lexists(self._attachment):
                raise OrmAttachmentNotExists(self._attachment)
            os.link(self._attachment, dst)

    def rename(self, new_filename):
        # TODO!
        pass

    @property
    def rel_path(self):
        if not self.is_attached:
            return None
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        if self._need_write:
            return None
        return c.storage.get_relative_filepath(self._instance, self._attr_name, self._value)

    @property
    def abs_path(self):
        if not self.is_attached:
            return None
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        if self._need_write:
            return self._attachment
        return c.storage.get_absolute_filepath(self._instance, self._attr_name, self._value)

    @property
    def filename(self):
        return self._value


class ValueFileAttachment(ValueAttachmentAbstract):
    pass


class ValuePictureAttachment(ValueAttachmentAbstract):
    def __init__(self, value):
        super(ValuePictureAttachment, self).__init__(value)

    def _push_on_write_image(self, f, src, filename, kw):
        fn, fp, img = f(src, self._instance, self._attr_name, filename, **kw)
        self._value = fn
        self.instance.__changes__[self.model_k] = ("%s '%s'" % (gettext('Attached', 'attachments'), fn), None)
        if self._instance != self.instance:
            self._instance.__changes__[self._attr_name] = ("%s '%s'" % (gettext('Attached', 'attachments'), fn), None)
        if self.attributes_instance is not None:
            img_width, img_height = img.size
            setattr(self.attributes_instance, 'filesize', int(math.ceil(float(os.stat(fp).st_size) / 1024.0)))
            setattr(self.attributes_instance, 'dimensions', (img_width, img_height))
            setattr(self.attributes_instance, 'format', str(img.format))

    def remove_image(self):
        self.remove_file()

    def push_image(self, image, filename=None, **kwargs):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._value = c.storage.suggest_filename(self._instance, self._attr_name, filename)
        self._on_write = self._push_on_write_image, (c.storage.push_image, image, str(self._value), kwargs)
        self._need_write = True
        self._attachment = image
        self.instance.__changes__[self.model_k] = (gettext('Attached', 'attachments'), None)
        if self._instance != self.instance:
            self._instance.__changes__[self._attr_name] = (gettext('Attached', 'attachments'), None)

    def push_image_from(self, src, filename=None, **kwargs):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._value = c.storage.suggest_filename(self._instance, self._attr_name, filename)
        self._on_write = self._push_on_write_image, (c.storage.push_image_from, src, str(self._value), kwargs)
        self._need_write = True
        self._attachment = src
        self.instance.__changes__[self.model_k] = (gettext('Attached', 'attachments'), None)
        if self._instance != self.instance:
            self._instance.__changes__[self._attr_name] = (gettext('Attached', 'attachments'), None)

    def pull_image_to(self, dst):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._check_attached()
        if dst[-1] == '/':
            dst += self._value
        if not self._need_write:
            c.storage.pull_image_to(dst, self._instance, self._attr_name)
        else:
            if isinstance(self._attachment, str):
                if not os.path.lexists(self._attachment):
                    raise OrmAttachmentNotExists(self._attachment)
                shutil.copyfile(self._attachment, dst)
            else:
                image_format = getattr(c.storage, 'image_format', None)
                quality = getattr(c.storage, 'quality', 100)
                if image_format is not None:
                    self._attachment.save(dst, image_format, quality=quality)
                else:
                    self._attachment.save(dst)

    def pull_image(self):
        from PIL import Image
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot deal with attachments until model been defined!")
        self._check_attached()
        if not self._need_write:
            return c.storage.pull_image(self._instance, self._attr_name)
        else:
            if isinstance(self._attachment, str):
                if not os.path.lexists(self._attachment):
                    raise OrmAttachmentNotExists(self._attachment)
                return Image.open(self._attachment)
            else:
                return self._attachment




import os.path
import uuid
import shutil
from .utils import remove_trailing_slash, fit_image
from .exceptions import OrmParameterError, OrmAttachmentNotExists, OrmAttachmentSourceNotExists, \
    OrmAttachmentAlreadyExists


class OrmStorageAbstract(object):
    """ A class used to define an actual files' placement - a storage. Storages are used to store
    files which beign attached to the objects outside database (as recommended behaviour). Files,
    which defines as a part of database object, are actually not besides in the database tables
    or its space, but instead they lays in the file system of the machine where this ORM
    is used at. """
    def __init__(self,
                 base_path,
                 with_model=True,
                 with_objpk=True,
                 with_attrname=True,
                 uuid_names=False,
                 prefix=None,
                 suffix=None,
                 allowed_ext=None,
                 prohibited_ext=None,
                 autoname=True):
        if not isinstance(base_path, str):
            raise OrmParameterError("storage 'base_path' must be a type of (str)!")
        if prefix is not None and not isinstance(prefix, str):
            raise OrmParameterError("storage's 'prefix' must be a type of (str) or not set!")
        if suffix is not None and not isinstance(suffix, str):
            raise OrmParameterError("storage's 'suffix' must be a type of (str) or not set!")
        self._base_path = remove_trailing_slash(base_path)
        self.with_model = bool(with_model)
        self.with_objpk = bool(with_objpk)
        self.with_attrname = bool(with_attrname)
        self.uuid_names = bool(uuid_names)
        self.prefix = prefix
        self.suffix = suffix
        self.allowed_ext = allowed_ext
        self.prohibited_ext = prohibited_ext
        self.autoname = bool(autoname)

    @property
    def base_path(self):
        """
        :returns                    Base file system path of the storage.
        """
        return self._base_path

    @base_path.setter
    def base_path(self, value):
        """ Sets the base file system path of the storage. """
        if not isinstance(value, str):
            raise OrmParameterError("storage 'base_path' must be a type of (str)!")
        self._base_path = remove_trailing_slash(value)

    def _compose_rel_path(self, instance, attr_name):
        d = list()
        if self.with_model:
            d.append(instance.__class__.__name__)
        if self.with_objpk:
            pk = instance.__class__.__meta__.primary_key
            vpk = list()
            for k in pk:
                vpk.append(str(getattr(instance, k)))
            d.append("@".join(vpk))
        if self.with_attrname:
            d.append(attr_name)
        return "/".join(d)

    def _compose_abs_path(self, instance, attr_name):
        return "/".join((self.base_path, self._compose_rel_path(instance, attr_name)))

    @staticmethod
    def _get_filename_from_attribute(instance, attr_name):
        # TODO ! special value classes for attachments
        # if self.uuid_names:
        #     return uuid.uuid4().hex
        return getattr(instance, attr_name).value

    def make_filename(self, instance, attr_name, filename=None, overwrite=False):
        filename = filename \
                   or (self._get_filename_from_attribute(instance, attr_name)
                       if not self.uuid_names
                       else uuid.uuid4().hex)
        if overwrite:
            return filename
        dirpath = self._compose_abs_path(instance, attr_name)
        f_name, f_ext = os.path.splitext(filename)
        filepath = "%s%s" % (f_name, f_ext)
        i = 1
        if os.path.lexists("%s/%s" % (dirpath, filepath)) and not self.autoname:
            raise OrmAttachmentAlreadyExists(
                "attachment file with name '%s' is already exists and 'autoname' set to False!" % filepath
            )
        while os.path.lexists("%s/%s" % (dirpath, filepath)):
            i += 1
            filepath = "%s (%i)%s" % (f_name, i, f_ext)
        return filepath

    def suggest_filename(self, instance, attr_name, filename=None, overwrite=False):
        if isinstance(filename, str) and '/' in filename:
            filename = os.path.basename(filename)
        filename = uuid.uuid4().hex \
            if self.uuid_names \
            else (filename if filename else self._get_filename_from_attribute(instance, attr_name))
        if overwrite:
            return filename
        dirpath = self._compose_abs_path(instance, attr_name)
        f_name, f_ext = os.path.splitext(filename)
        filepath = "%s%s" % (f_name, f_ext)
        i = 1
        if os.path.lexists("%s/%s" % (dirpath, filepath)) and not self.autoname:
            raise OrmAttachmentAlreadyExists(
                "attachment file with name '%s' is already exists and 'autoname' set to False!" % filepath
            )
        while os.path.lexists("%s/%s" % (dirpath, filepath)):
            i += 1
            filepath = "%s (%i)%s" % (f_name, i, f_ext)
        return filepath

    def copy_file_from(self, source, instance, attr_name, filename=None):
        """ Puts the file into the storage. File must exists. If filename is omitted - it
        will be taken from the source filename.
        This method copies the file from the original source, not removing it's source. This might
        take a time to complete, depending on the size of the copying attachment file.
        :source                 Full path to the file which to put to the storage;
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               Filename to set up (not required to be the same as original
                                filename of the file); if omitted - the filename of the
                                original file will be taken;
        :returns                A tuple consisting of resulting filename and path to it.
        """
        if not filename:
            filename = os.path.basename(source)
        filename = self.make_filename(instance, attr_name, filename)
        attachment_dir = self._compose_abs_path(instance, attr_name)
        if not os.path.isdir(attachment_dir):
            os.makedirs(attachment_dir)
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if os.path.lexists(attachment_path):
            raise OrmAttachmentAlreadyExists(attachment_path)
        if not os.path.lexists(source):
            raise OrmAttachmentSourceNotExists(source)
        shutil.copyfile(source, attachment_path)
        return filename, attachment_path

    def link_file_from(self, source, instance, attr_name, filename=None):
        """ Puts the file into the storage. File must exists. If filename is omitted - it
        will be taken from the source filename.
        This method makes a hard link from the source file to the storage. Source file not
        removes, but both source file and attachment will point to the same block of file
        system, so any changes in the content of any of them will affect the other too.
        Note that this operation requires that both source file and attachment (storage)
        to be located in the same file system (mounted FS).
        :source                 Full path to the file which to put to the storage;
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               Filename to set up (not required to be the same as original
                                filename of the file); if omitted - the filename of the
                                original file will be taken;
        :returns                A tuple consisting of resulting filename and path to it.
        """
        if not filename:
            filename = os.path.basename(source)
        filename = self.make_filename(instance, attr_name, filename)
        attachment_dir = self._compose_abs_path(instance, attr_name)
        if not os.path.isdir(attachment_dir):
            os.makedirs(attachment_dir)
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if os.path.lexists(attachment_path):
            raise OrmAttachmentAlreadyExists(attachment_path)
        if not os.path.lexists(source):
            raise OrmAttachmentSourceNotExists(source)
        os.link(source, attachment_path)
        return filename, attachment_path

    def move_file_from(self, source, instance, attr_name, filename=None):
        """ Puts the file into the storage. File must exists. If filename is omitted - it
        will be taken from the source filename.
        This method moves file from the original file system position to the storage. Most
        often this is the preferred option because it is fast while source file and the
        storage are located at the same file system (mounted FS).
        :source                 Full path to the file which to put to the storage;
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               Filename to set up (not required to be the same as original
                                filename of the file); if omitted - the filename of the
                                original file will be taken;
        :returns                A tuple consisting of resulting filename and path to it.
        """
        if not filename:
            filename = os.path.basename(source)
        filename = self.make_filename(instance, attr_name, filename)
        attachment_dir = self._compose_abs_path(instance, attr_name)
        if not os.path.isdir(attachment_dir):
            os.makedirs(attachment_dir)
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if os.path.lexists(attachment_path):
            raise OrmAttachmentAlreadyExists(attachment_path)
        if not os.path.lexists(source):
            raise OrmAttachmentSourceNotExists(source)
        shutil.move(source, attachment_path)
        return filename, attachment_path

    def move_file_to(self, destination, instance, attr_name, filename=None):
        """ Moves the file from the storage to the specified destination position of the
        file system. File becomes detached from the object instance after.
        :destination            Full path, with filename, where to store a file from the
                                storage;
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               The filename of the file in the storage. If not set - the
                                method will get filename from the instance's attribute's
                                value (it's representation of the attachment). Usually this
                                is an only right way.
        """
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if not os.path.lexists(attachment_path):
            raise OrmAttachmentNotExists(attachment_path)
        shutil.move(attachment_path, destination)

    def copy_file_to(self, destination, instance, attr_name, filename=None):
        """ Copies the file from the storage to the specified destination position of the
        file system. File keeps attached after. This operation may take a time to complete
        because file system has to actually copy file from storage.
        :destination            Full path, with filename, where to store a file from the
                                storage;
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               The filename of the file in the storage. If not set - the
                                method will get filename from the instance's attribute's
                                value (it's representation of the attachment). Usually this
                                is an only right way.
        """
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if not os.path.lexists(attachment_path):
            raise OrmAttachmentNotExists(attachment_path)
        shutil.copyfile(attachment_path, destination)

    def link_file_to(self, destination, instance, attr_name, filename=None):
        """ Creates a hard link from attached file to the destination one. This operation
        not detachs file from the object instance, but creates a pointer to the same file
        system blocks in the specified position, so any changes in the content of any one
        of those two files will affect the other too.
        Note that this operation requires that both destination file and attachment (storage)
        to be located in the same file system (mounted FS).
        :destination            Full path, with filename, where to store a file from the
                                storage;
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               The filename of the file in the storage. If not set - the
                                method will get filename from the instance's attribute's
                                value (it's representation of the attachment). Usually this
                                is an only right way.
        """
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if not os.path.lexists(attachment_path):
            raise OrmAttachmentNotExists(attachment_path)
        os.link(attachment_path, destination)

    def del_file(self, instance, attr_name, filename=None):
        """ Removes the file from the storage.
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               The filename of the file in the storage. If not set - the
                                method will get filename from the instance's attribute's
                                value (it's representation of the attachment). Usually this
                                is an only right way.
        """
        filepath = self.get_absolute_filepath(instance, attr_name, filename)
        if not os.path.lexists(filepath):
            return
        os.unlink(filepath)

    def rename_file(self, new_filename, instance, attr_name, filename=None):
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if not os.path.lexists(attachment_path):
            raise OrmAttachmentNotExists(attachment_path)
        new_filename = self.make_filename(instance, attr_name, new_filename)
        new_attachment_path = self.get_absolute_filepath(instance, attr_name, new_filename)
        shutil.move(attachment_path, new_attachment_path)
        return new_filename, new_attachment_path

    def verify_file_existance(self, instance, attr_name, filename=None):
        """ Verify that specified file is really exists in the storage. Useful when object's
        attribute tells that file is attached, but ORM mechanics wants to verify that (because
        database can tell us that file is attached, but it might been deleted).
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               The filename of the file in the storage. If not set - the
                                method will get filename from the instance's attribute's
                                value (it's representation of the attachment). Usually this
                                is an only right way.
        :returns                True if the file is exists or filename cannot be resolved;
                                False if the file is not exists physically.
        """
        filename = filename or self._get_filename_from_attribute(instance, attr_name)
        if not filename:
            return True
        return os.path.exists(self.get_absolute_filepath(instance, attr_name, filename))

    def get_relative_filepath(self, instance, attr_name, filename=None):
        """ Returns, relative to the storage base path, file path.
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               The filename of the file in the storage. If not set - the
                                method will get filename from the instance's attribute's
                                value (it's representation of the attachment). Usually this
                                is an only right way.
        :returns                Relative path to the attached file.
        """
        filename = filename or self._get_filename_from_attribute(instance, attr_name)
        return "/".join((self._compose_rel_path(instance, attr_name), filename))

    def get_absolute_filepath(self, instance, attr_name, filename=None):
        """ Returns absolute file path in the server's file system.
        :instance               The object instance which holds this file attachment;
        :attr_name              The name of the attribute of the given object instance which
                                represents the file attachment;
        :filename               The filename of the file in the storage. If not set - the
                                method will get filename from the instance's attribute's
                                value (it's representation of the attachment). Usually this
                                is an only right way.
        :returns                Absolute path to the attached file.
        """
        return "/".join((self.base_path, self.get_relative_filepath(instance, attr_name, filename)))


class OrmFileStorage(OrmStorageAbstract):
    """ General files storage. """
    pass


class OrmPictureStorage(OrmStorageAbstract):
    """ Picture storage which implements additional methods to work with images and additional
    logics for pictures' placing. """

    def __init__(self, base_path, **kwargs):
        self.image_format = kwargs.pop('image_format', None)
        self.max_width = kwargs.pop('max_width', None)
        self.max_height = kwargs.pop('max_height', None)
        self.min_width = kwargs.pop('min_width', None)
        self.min_height = kwargs.pop('min_height', None)
        self.quality = kwargs.pop('quality', 100)
        if self.max_width is not None and not isinstance(self.max_width, int):
            raise OrmParameterError("picture storage requires 'max_width' to be type (int) or not set!")
        if self.max_height is not None and not isinstance(self.max_height, int):
            raise OrmParameterError("picture storage requires 'max_height' to be type (int) or not set!")
        if self.min_width is not None and not isinstance(self.min_width, int):
            raise OrmParameterError("picture storage requires 'min_width' to be type (int) or not set!")
        if self.min_height is not None and not isinstance(self.min_height, int):
            raise OrmParameterError("picture storage requires 'min_height' to be type (int) or not set!")
        if not isinstance(self.quality, int) or 0 >= self.quality > 100:
            raise OrmParameterError("picture storage requires 'quality' to be type (int) and be in [1..100]!")
        super(OrmPictureStorage, self).__init__(base_path, **kwargs)

    def _has_dim_limits(self, **kwargs):
        return bool(self.max_height
                    or self.max_width
                    or self.min_height
                    or self.min_width
                    or kwargs.get('max_height', None)
                    or kwargs.get('max_width', None)
                    or kwargs.get('min_height', None)
                    or kwargs.get('min_width', None))

    def _ensure_image_configuration(self, image, **kwargs):
        if not self._has_dim_limits(**kwargs):
            return
        max_width = kwargs.get('max_width', None) or self.max_width
        max_height = kwargs.get('max_height', None) or self.max_height
        min_width = kwargs.get('min_width', None) or self.min_width
        min_height = kwargs.get('min_height', None) or self.min_height
        fit_image(image, max_height=max_height, min_height=min_height, max_width=max_width, min_width=min_width)

    def push_image(self, image, instance, attr_name, filename=None, **kwargs):
        if not filename and not self.uuid_names:
            raise OrmParameterError(
                "picture storage requires filename to be given on 'push_image' when 'uuid_names' is False!"
            )
        filename = self.make_filename(instance, attr_name, filename)
        attachment_dir = self._compose_abs_path(instance, attr_name)
        if not os.path.isdir(attachment_dir):
            os.makedirs(attachment_dir)
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if os.path.lexists(attachment_path):
            raise OrmAttachmentAlreadyExists(attachment_path)
        image_format = kwargs.get('image_format', None) or self.image_format
        quality = kwargs.get('quality', 100)
        self._ensure_image_configuration(image, **kwargs)
        if image_format is not None:
            image.save(attachment_path, image_format, quality=quality)
        else:
            image.save(attachment_path)
        return filename, attachment_path, image

    def push_image_from(self, source, instance, attr_name, filename=None, **kwargs):
        from PIL import Image
        if not filename and not self.uuid_names:
            raise OrmParameterError(
                "picture storage requires filename to be given on 'push_image' when 'uuid_names' is False!"
            )
        filename = self.make_filename(instance, attr_name, filename)
        attachment_dir = self._compose_abs_path(instance, attr_name)
        if not os.path.isdir(attachment_dir):
            os.makedirs(attachment_dir)
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if os.path.lexists(attachment_path):
            raise OrmAttachmentAlreadyExists(attachment_path)
        if not os.path.lexists(source):
            raise OrmAttachmentSourceNotExists(source)
        image = Image.open(source)
        image_format = kwargs.get('image_format', None) or self.image_format
        quality = kwargs.get('quality', 100)
        self._ensure_image_configuration(image, **kwargs)
        if image_format is not None:
            image.save(attachment_path, image_format, quality=quality)
        else:
            image.save(attachment_path)
        return filename, attachment_path, image

    def pull_image_to(self, destination, instance, attr_name, filename=None):
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if not os.path.lexists(attachment_path):
            raise OrmAttachmentNotExists(attachment_path)
        shutil.copyfile(attachment_path, destination)

    def pull_image(self, instance, attr_name, filename=None):
        from PIL import Image
        attachment_path = self.get_absolute_filepath(instance, attr_name, filename)
        if not os.path.lexists(attachment_path):
            raise OrmAttachmentNotExists(attachment_path)
        return Image.open(attachment_path)




import uuid
import re
import datetime
import hashlib
from .base import OrmComparator
from .columns import ColumnIntegerAbstract, ColumnFloatAbstract, ColumnNumericAbstract, ColumnDatetimeAbstract, \
    ColumnDateAbstract, ColumnTimeAbstract, ColumnStringAbstract, ColumnTextAbstract, ColumnBlobAbstract, \
    ColumnBooleanAbstract, ColumnAttributeAbstract
from ..values import AttributeValueNull, AttributeValueString, ValueYearMonth, ValuePhone, ValueIPRequesties, \
    ValueDimensions2D, ValueDimensions3D, ValuePoint, ValueBox, ValueVector, ValueGeoCoordinate, ValuePassword
from ..funcs import if_, uncastable_, inet6_aton_, inet6_ntoa_, host_, concat_ws_, substr_, concat_, unhex_, length_, \
    conv_, hex_, pg_text_
from ..utils import gettext, ipaddr_version, ipv4mask_2_cidr
from ..manage import manager
from ..exceptions import OrmValueVerificationError, OrmPasswordIsWeak, OrmProhibittedValueSet, OrmValueError, \
    OrmDefinitionError, OrmParameterError


PASSWORD_ALGORYTHM_PLAIN = 'plain'
PASSWORD_ALGORYTHM_MD5 = 'md5'
PASSWORD_ALGORYTHM_SHA1 = 'sha1'
PASSWORD_ALGORYTHM_SHA224 = 'sha224'
PASSWORD_ALGORYTHM_SHA256 = 'sha256'
PASSWORD_ALGORYTHM_SHA384 = 'sha384'
PASSWORD_ALGORYTHM_SHA512 = 'sha512'


class Integer(ColumnIntegerAbstract):
    def __init__(self, **kwargs):
        kwargs['big'] = False
        super(Integer, self).__init__(11, **kwargs)


class BigInteger(ColumnIntegerAbstract):
    def __init__(self, **kwargs):
        kwargs['big'] = True
        super(BigInteger, self).__init__(20, **kwargs)


class PositiveInteger(ColumnIntegerAbstract):
    def __init__(self, **kwargs):
        kwargs['big'] = False
        kwargs['unsigned'] = True
        super(PositiveInteger, self).__init__(11, **kwargs)


class PositiveBigInteger(ColumnIntegerAbstract):
    def __init__(self, **kwargs):
        kwargs['big'] = True
        kwargs['unsigned'] = True
        super(PositiveBigInteger, self).__init__(20, **kwargs)


class Float(ColumnFloatAbstract):
    def __init__(self, length=None, decimals=None, **kwargs):
        kwargs['big'] = False
        super(Float, self).__init__(length, decimals, **kwargs)


class BigFloat(ColumnFloatAbstract):
    def __init__(self, length=None, decimals=None, **kwargs):
        kwargs['big'] = True
        super(BigFloat, self).__init__(length, decimals, **kwargs)


class PositiveFloat(ColumnFloatAbstract):
    def __init__(self, length=None, decimals=None, **kwargs):
        kwargs['big'] = False
        kwargs['unsigned'] = True
        super(PositiveFloat, self).__init__(length, decimals, **kwargs)


class PositiveBigFloat(ColumnFloatAbstract):
    def __init__(self, length=None, decimals=None, **kwargs):
        kwargs['big'] = True
        kwargs['unsigned'] = True
        super(PositiveBigFloat, self).__init__(length, decimals, **kwargs)


class Numeric(ColumnNumericAbstract):
    pass


class Currency(ColumnNumericAbstract):
    def __init__(self, currency=None, currency_format=None, **kwargs):
        self.always_with_decimal = kwargs.pop('always_with_decimal', True)
        super(Currency, self).__init__(**kwargs)
        self._currency = currency
        self._format = currency_format

    @property
    def currency(self):
        return self._currency

    @currency.setter
    def currency(self, value):
        self._currency = value or manager.get_default_currency()
        self._format = manager.get_currency_format(self._currency) if self._currency is not None else None

    @property
    def currency_format(self):
        return self._format

    @currency_format.setter
    def currency_format(self, value):
        if value is None:
            self._format = manager.get_currency_format(self._currency) if self._currency is not None else None
        if not isinstance(value, (list, tuple)) or len(value) != 2:
            raise OrmValueError("currency format must be set using tuple of full and integer representations!")
        self._format = value

    @property
    def full_format(self):
        return "%i.%d" if self._format is None else self._format[0]

    @property
    def integer_format(self):
        return "%i" if self._format is None else self._format[1]

    def text_repr(self, value, instance=None):
        if value is None:
            return ""
        if not isinstance(value, (int, float)):
            raise OrmValueError("currency expects (float|int|None) type of value to be given!")
        value = float(value)
        use_full_format = self.always_with_decimal or not value.is_integer()
        format_ = self.full_format if use_full_format else self.integer_format
        i, d = divmod(value, 1)
        return format_.replace("%i", str(int(round(i)))).replace("%d", "%02d" % str(int(round(d * 100))))


class PositiveNumeric(ColumnNumericAbstract):
    def __init__(self, length, decimals, **kwargs):
        kwargs['unsigned'] = True
        super(PositiveNumeric, self).__init__(length, decimals, **kwargs)


class String(ColumnStringAbstract):
    pass


class Label(String):
    def __init__(self, length=250, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', False)
        kwargs['default'] = kwargs.get('default', '')
        kwargs['required'] = True
        kwargs['max_chars'] = length
        kwargs['findable'] = True
        super(Label, self).__init__(length, **kwargs)


class CaptionString(String):
    def __init__(self, length=60, caption=None, abbreviation=None, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', False)
        kwargs['default'] = kwargs.get('default', '')
        kwargs['required'] = True
        kwargs['is_caption_key'] = True
        kwargs['max_chars'] = length
        kwargs['findable'] = True
        kwargs['caption'] = caption or kwargs.get('caption', None)
        kwargs['abbreviation'] = abbreviation or kwargs.get('abbreviation', None)
        super(CaptionString, self).__init__(length, **kwargs)


class Uuid(String):
    """ UUID data type to store unique values, basing on Python's UUID4 function. """

    def __init__(self, **kwargs):
        kwargs['strict_find'] = True
        super(Uuid, self).__init__(36, **kwargs)

    @staticmethod
    def make_uuid():
        return str(uuid.uuid4())

    def verify(self, value):
        value = super(Uuid, self).verify(value)
        if value is None:
            return None
        value = str(value).replace('-', '').lower()
        if re.sub("[^0-9a-f]", "", value) != value:
            raise OrmValueVerificationError(gettext("Must be valid UUID"))
        if len(value) != 32:
            raise OrmValueVerificationError(gettext("Must be valid UUID"))
        return "%s-%s-%s-%s-%s" % (value[0:8], value[8:12], value[12:16], value[16:20], value[20:])


class Text(ColumnTextAbstract):
    pass


class LongText(ColumnTextAbstract):
    def __init__(self, **kwargs):
        kwargs['big'] = True
        super(LongText, self).__init__(**kwargs)


class Boolean(ColumnBooleanAbstract):
    @staticmethod
    def _getter(instance, key):
        value = instance.__getattrabs__(key)
        if isinstance(value, str):
            return value
        # return None if value is None else bool(value)
        return value

    def __init__(self, **kwargs):
        kwargs['getter'] = kwargs.get('getter', self._getter)
        kwargs['nullable'] = kwargs.get('nullable', False)
        kwargs['options'] = {
            1: kwargs.get('caption_yes', None) or gettext('Yes'),
            0: kwargs.get('caption_no', None) or gettext('No')
        }
        super(Boolean, self).__init__(**kwargs)

    def verify(self, value):
        value = super(Boolean, self).verify(value)
        if value is None:
            return None
        if isinstance(value, str) \
                and (value.lower() in (
                    gettext('No').lower(),
                    gettext('Off').lower(),
                    gettext('Disable').lower(),
                    gettext('False').lower(),
                    'n',
                    'f',
                ) or value.lower()[0] in ('n', 'f')):
            return False
        return bool(value)


class Date(ColumnDateAbstract):
    def __init__(self, **kwargs):
        super(Date, self).__init__(**kwargs)


class Time(ColumnTimeAbstract):
    def __init__(self, fsp=None, **kwargs):
        self._hasdate = False
        self._hastime = True
        super(Time, self).__init__(fsp, **kwargs)


class DateTime(ColumnDatetimeAbstract):
    def __init__(self, fsp=None, **kwargs):
        self._hasdate = True
        self._hastime = True
        super(DateTime, self).__init__(fsp, **kwargs)


class Blob(ColumnBlobAbstract):
    def __init__(self, **kwargs):
        kwargs['big'] = False
        super(Blob, self).__init__(**kwargs)


class LongBlob(ColumnBlobAbstract):
    def __init__(self, **kwargs):
        kwargs['big'] = True
        super(LongBlob, self).__init__(**kwargs)


class UuidPrimaryKey(Uuid):
    def __init__(self, **kwargs):
        kwargs['primary_key'] = True
        kwargs['nullable'] = False
        kwargs['default'] = self.make_uuid
        super(UuidPrimaryKey, self).__init__(**kwargs)
        self.base_type = Uuid


class AutoIncrement(PositiveInteger):
    def __init__(self, **kwargs):
        kwargs['auto_increment'] = True
        kwargs['primary_key'] = True
        kwargs['nullable'] = False
        super(AutoIncrement, self).__init__(**kwargs)
        self.base_type = PositiveInteger


class BigAutoIncrement(PositiveBigInteger):
    def __init__(self, **kwargs):
        kwargs['auto_increment'] = True
        kwargs['primary_key'] = True
        kwargs['nullable'] = False
        super(BigAutoIncrement, self).__init__(**kwargs)
        self.base_type = PositiveBigInteger


class ObjectVersion(PositiveInteger):
    def __init__(self, **kwargs):
        kwargs['is_version_key'] = True
        kwargs['default'] = kwargs.get('default', 1)
        kwargs['nullable'] = kwargs.get('nullable', False)
        super(ObjectVersion, self).__init__(**kwargs)


class ObjectVersionTimestamp(DateTime):
    def __init__(self, **kwargs):
        kwargs['is_version_key'] = True
        super(ObjectVersionTimestamp, self).__init__(**kwargs)

    def on_model_declared(self):
        self.model.__meta__.timestamp_modification.append(self)


class ObjectSoftdeleteAndVersion(Integer):
    def __init__(self, **kwargs):
        kwargs['is_version_key'] = True
        kwargs['default'] = kwargs.get('default', 1)
        kwargs['nullable'] = True
        kwargs['is_softdelete_key'] = True
        super(ObjectSoftdeleteAndVersion, self).__init__(**kwargs)


class SoftdeleteTimestamp(DateTime):
    def __init__(self, **kwargs):
        kwargs['is_softdelete_key'] = True
        super(SoftdeleteTimestamp, self).__init__(**kwargs)


class SoftdeleteBoolean(Boolean):
    def __init__(self, **kwargs):
        kwargs['is_softdelete_key'] = True
        super(SoftdeleteBoolean, self).__init__(**kwargs)


class SoftdeleteDate(Date):
    def __init__(self, **kwargs):
        kwargs['is_softdelete_key'] = True
        super(SoftdeleteDate, self).__init__(**kwargs)


class CreationTimestamp(DateTime):
    def __init__(self, **kwargs):
        kwargs['default'] = datetime.datetime.now
        super(CreationTimestamp, self).__init__(**kwargs)

    def on_model_declared(self):
        self.model.__meta__.timestamp_creation.append(self)


class ModificationTimestamp(DateTime):
    def __init__(self, **kwargs):
        kwargs['default'] = datetime.datetime.now
        super(ModificationTimestamp, self).__init__(**kwargs)

    def on_model_declared(self):
        self.model.__meta__.timestamp_modification.append(self)


class Password(String):
    @staticmethod
    def _is_password_strong(value):
        l_ = len(value)
        vn = ''.join(list((filter(lambda x: not x.isdigit(), value))))
        if len(vn) == l_ or len(vn) == 0:
            return False
        return True

    def _setter(self, instance, key, value, on_instance_loaded=False):
        if on_instance_loaded:
            instance.__setattrvalue__(key, value)
            return value
        if not value:
            if self.empty is False:
                raise OrmProhibittedValueSet()
            else:
                instance.__setattrvalue__(key, self.empty)
                return self.empty
        if self.strong and not self._is_password_strong(value):
            raise OrmPasswordIsWeak(
                "The password is about to set to the `Password` attribute is too weak: '%s'!" % str(value)
            )
        value += self.salt
        if self.encryption == 'plain':
            instance.__setattrvalue__(key, value)
            return value
        elif self.encryption == 'md5':
            value_ = hashlib.md5(value.encode('utf-8')).hexdigest().lower()
        elif self.encryption == 'sha1':
            value_ = hashlib.sha1(value.encode('utf-8')).hexdigest().lower()
        elif self.encryption == 'sha224':
            value_ = hashlib.sha224(value.encode('utf-8')).hexdigest().lower()
        elif self.encryption == 'sha256':
            value_ = hashlib.sha256(value.encode('utf-8')).hexdigest().lower()
        elif self.encryption == 'sha384':
            value_ = hashlib.sha384(value.encode('utf-8')).hexdigest().lower()
        elif self.encryption == 'sha512':
            value_ = hashlib.sha512(value.encode('utf-8')).hexdigest().lower()
        else:
            value_ = None
        instance.__setattrvalue__(key, value_)
        return value_

    @staticmethod
    def _text_repr(void):
        return "******"

    def __init__(self, *args, **kwargs):
        length = None
        encryption = None
        for arg in args:
            if isinstance(arg, int) and arg > 0:
                length = arg
            elif isinstance(arg, str) \
                    and arg.lower() in ('plain', 'md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512'):
                encryption = arg.lower()
            else:
                AttributeError("Unknown value given for `Password`: %s" % str(arg))
        encryption = str(kwargs.get('encryption', 'sha256')).lower() if encryption is None else encryption
        if encryption not in ('plain', 'md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512'):
            raise AttributeError(
                "`Password` attribute supports only next encryption algorithms:"
                " 'plain', 'md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512'!"
            )
        length = kwargs.get('length', None) if length is None else length
        salt = str(kwargs.get('salt')) if kwargs.get('salt', None) else ''
        empty = kwargs.get('empty', False)
        if empty not in (False, None, ''):
            raise AttributeError("`Password`.`empty` mght be False, None or empty string ('') only!")
        strong = bool(kwargs.get('strong', False))
        if length and length > 255:
            raise AttributeError("`Password` attribute supports maximum 255 chars length of column!")
        if length is None:
            if encryption == 'plain':
                length = 255
            elif encryption == 'md5':
                length = 32
            elif encryption == 'sha1':
                length = 40
            elif encryption == 'sha224':
                length = 56
            elif encryption == 'sha256':
                length = 64
            elif encryption == 'sha384':
                length = 96
            elif encryption == 'sha512':
                length = 128
        nullable = empty is not False
        default = None if nullable else ''
        kwargs['nullable'] = nullable
        kwargs['default'] = default
        kwargs['listable'] = False
        kwargs['findable'] = False
        kwargs['setter'] = self._setter
        kwargs['on_text_repr'] = self._text_repr
        kwargs['on_log_repr'] = str(gettext('Changed'))
        super(Password, self).__init__(length, **kwargs)
        self.value_class = ValuePassword
        self.salt = salt
        self.empty = empty
        self.strong = strong
        self.encryption = encryption
        if encryption == 'plain':
            self.collate = 'utf8_bin'

    def corresponds_to_(self, plain):
        """ Used in SELECT requests, comparing given 'plain' textual password with corresponding
        hashed value stored in the database. """
        if plain is None or isinstance(plain, AttributeValueNull):
            return OrmComparator(self, 'ISNULL')
        if self.encryption == 'plain':
            return self.__eq__(plain)
        elif self.encryption == 'md5':
            return self.__eq__(hashlib.md5(plain.encode('utf-8')).hexdigest().lower())
        elif self.encryption == 'sha1':
            return self.__eq__(hashlib.sha1(plain.encode('utf-8')).hexdigest().lower())
        elif self.encryption == 'sha224':
            return self.__eq__(hashlib.sha224(plain.encode('utf-8')).hexdigest().lower())
        elif self.encryption == 'sha256':
            return self.__eq__(hashlib.sha256(plain.encode('utf-8')).hexdigest().lower())
        elif self.encryption == 'sha384':
            return self.__eq__(hashlib.sha384(plain.encode('utf-8')).hexdigest().lower())
        elif self.encryption == 'sha512':
            return self.__eq__(hashlib.sha512(plain.encode('utf-8')).hexdigest().lower())
        return self.__eq__(plain)

    def try_password(self, password):  # Abstract (->ValuePassword)
        pass


class YearMonth(Float):
    def __init__(self, **kwargs):
        super(YearMonth, self).__init__(**kwargs)
        self.enumerable_value = True
        self.value_class = ValueYearMonth
        self.year_first = bool(kwargs.get('year_first', True))
        self.separator = str(kwargs.get('separator', '.'))


class Phone(String):
    def __init__(self,
                 formats=None,
                 format_ext=None,
                 local_prefixes=None,
                 store_local_prefix=True,
                 **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', False)
        kwargs['default'] = kwargs.get('default', '')
        if formats is not None and not isinstance(formats, (list, tuple)):
            raise OrmDefinitionError("Phone attribute's options 'formats' must be a type of list or tuple!")
        elif formats:
            formats_ = dict()
            for f in formats:
                fl = len(f) - len(f.replace('?', ''))
                formats_[fl] = f
        else:
            formats_ = manager.get_telephony_formats()
        if local_prefixes and not isinstance(local_prefixes, dict):
            raise OrmDefinitionError(
                "Phone attribute option 'local_prefixes' must be a type of dict in format"
                " ({<base_length>: <prefix>, <base_length>: <prefix>})"
            )
        super(Phone, self).__init__(63, **kwargs)
        self.enumerable_value = True
        self.value_class = ValuePhone
        self.formats = formats_
        self.format_ext = format_ext if format_ext else manager.get_telephony_format_ext()
        self.local_prefixes = local_prefixes if local_prefixes else manager.get_telephony_local_prefixes()
        self.store_local_prefix = bool(store_local_prefix)

    def _apply_local_prefix(self, phone):
        if phone is None:
            return None
        plen = len(phone)
        if plen in self.local_prefixes:
            phone = str(self.local_prefixes[plen]) + phone
        return phone

    def do_store_local_prefix(self, phone):
        if phone is None:
            return None
        if not self.store_local_prefix:
            return phone
        plen = len(phone)
        if plen in self.local_prefixes:
            phone = str(self.local_prefixes[plen]) + phone
        return phone

    def format_phone(self, phone, ext):
        if not phone:
            phone = ""
        phone = re.sub("[^0-9]", "", str(phone))
        if phone:
            phone = self._apply_local_prefix(phone)
        plen = len(phone)
        if plen in self.formats:
            fv = self.formats[plen]
            for ch in phone:
                fv = fv.replace('?', ch, 1)
            phone = fv
        if ext:
            ext_ = self.format_ext.replace('?', re.sub("[^0-9]", "", ext))
            phone = "%s%s" % (phone, ext_)
        return phone

    @staticmethod
    def _decompose_phone(value):
        if not value:
            return '', ''
        if ',' in value:
            phone_ext = value.split(',')
            phone = phone_ext[0]
            ext = phone_ext[1]
        else:
            phone = value
            ext = ''
        return phone, ext

    @staticmethod
    def _find_prepare(value):
        if not value:
            return value
        return re.sub("[^0-9]", "", str(value))


class Email(String):
    def __init__(self, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', False)
        kwargs['default'] = kwargs.get('default', '')
        super(Email, self).__init__(60, **kwargs)

    def verify(self, value):
        value = super(Email, self).verify(value)
        if value is None:
            return None
        if not self.required and not value:
            return ""
        value = str(value).lower()
        if '@' not in value or value.count('@') > 1:
            raise OrmValueVerificationError(gettext("Must be an e-mail address"))
        if re.sub("[^0-9a-z.@\-_+\[\]]", "", value) != value:
            raise OrmValueVerificationError(gettext("Must be an e-mail address"))
        p = value.split('@')
        if not p[0] or not p[1]:
            raise OrmValueVerificationError(gettext("Must be an e-mail address"))
        if '.' not in p[1] or not ('[' in p[1] and ']' in p[1]):
            raise OrmValueVerificationError(gettext("Must be an e-mail address"))
        return value


class IMAddress(String):
    def __init__(self, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', False)
        kwargs['default'] = kwargs.get('default', '')
        super(IMAddress, self).__init__(120, **kwargs)


class SocialNetworkAddress(String):
    def __init__(self, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', False)
        kwargs['default'] = kwargs.get('default', '')
        super(SocialNetworkAddress, self).__init__(250, **kwargs)


class IPAddress(ColumnAttributeAbstract):
    def __init__(self, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', True)
        kwargs['default'] = kwargs.get('default', None)
        kwargs['cast_write'] = self._cast_write
        kwargs['cast_read'] = self._cast_read
        kwargs['cast_order_by'] = None
        kwargs['cast_group_by'] = self._cast_read
        kwargs['strict_find'] = True
        super(IPAddress, self).__init__(**kwargs)
        self.value_class = AttributeValueString

    @staticmethod
    def _cast_write(context, pn):
        if context.session.dbms_type == 'mysql' and context.params[pn] is not None:
            return inet6_aton_(context.params[pn])
        return "%(" + pn + ")s"

    def _cast_read(self, context):
        if context.session.dbms_type == 'mysql':
            return inet6_ntoa_(uncastable_(self))
        elif context.session.dbms_type == 'postgresql':
            return host_(uncastable_(self))
        return self

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() != 'varbinary':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if struct['character_maximum_length'] != 32:
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['data_type']).lower() != 'inet':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        if session.dbms_type == 'mysql':
            d.append("VARBINARY(32)")
        elif session.dbms_type == 'postgresql':
            d.append("INET")
        else:
            raise TypeError("this column type is not supported for this database and has no alternative!")
        if not self.nullable:
            d.append('NOT NULL')
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            d.append("SET DATA TYPE INET")
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def verify(self, value):
        value = super(IPAddress, self).verify(value)
        if value is None:
            return None
        if not self.required and not value:
            return ""
        value = str(value)
        ipvers = ipaddr_version(value)
        if ipvers == 0:
            raise OrmValueVerificationError(gettext("Must be an IP address"))
        return value

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))

    @property
    def notempty_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def true_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def false_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))


class IPMask(Integer):
    """ Network IP mask stored at the DBMS side as small integer value between 0 and 128 for
    IPv6, and 0 and 32 for IPv4 (conforming to CIDR specification). So, textual IPv4 mask
    representation as A.B.C.D will be converted to the numeric CIDR anyway, if given. """

    def __init__(self, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', True)
        kwargs['default'] = kwargs.get('default', None)
        kwargs['unsigned'] = True
        self.prefix_with_slash = bool(kwargs.pop('prefix_with_slash', False))
        super(IPMask, self).__init__(**kwargs)

    def verify(self, value):
        value = super(IPMask, self).verify(value)
        if value is None:
            return None
        if not isinstance(value, (int, str)):
            raise OrmValueError("IPMask can be set using (str|int) type only!")
        if isinstance(value, int):
            if value < 0 or value > 128:
                raise OrmValueVerificationError(gettext("Must be a network mask"))
            return value
        ipvers = ipaddr_version(value)
        if ipvers != 4:
            raise OrmValueVerificationError(gettext("Must be a network mask"))
        return int(ipv4mask_2_cidr(value))

    def text_repr(self, value, instance=None):
        if value is None:
            return ""
        return "/%i" % int(value) if self.prefix_with_slash else str(value)


class MACAddress(String):
    def __init__(self, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', True)
        kwargs['default'] = kwargs.get('default', None)
        kwargs['cast_write'] = self._cast_write
        kwargs['cast_read'] = self._cast_read
        kwargs['cast_order_by'] = None
        kwargs['cast_group_by'] = self._cast_read
        kwargs['strict_find'] = True
        self.word_sz = kwargs.pop('word_sz', 2)
        self.separator = kwargs.pop('separator', ':')
        super(MACAddress, self).__init__(16, **kwargs)
        self.value_handler = self._value_handler

    @staticmethod
    def _cast_write(context, pn):
        value = context.params[pn]
        if value is None:
            return "NULL"
        value = re.sub("[^0-9a-f]", "", str(value).lower())
        sz = len(value)
        if sz < 12:
            value = "%s%s" % ("0" * (12 - sz), value)
        return '"%s"' % str(value)

    def _cast_read(self, context):
        return concat_ws_(
            self.separator,
            substr_(uncastable_(self), 1, 4),
            substr_(uncastable_(self), 5, 4),
            substr_(uncastable_(self), 9, 4)
        ) if self.word_sz == 4 else concat_ws_(
            self.separator,
            substr_(uncastable_(self), 1, 2),
            substr_(uncastable_(self), 3, 2),
            substr_(uncastable_(self), 5, 2),
            substr_(uncastable_(self), 7, 2),
            substr_(uncastable_(self), 9, 2),
            substr_(uncastable_(self), 11, 2),
        )

    def _format_value(self, value):
        value = re.sub("[^0-9a-f]", "", str(value).lower())
        sz = len(value)
        if sz < 12:
            value = "%s%s" % ("0" * (12 - sz), value)
        return "%s%s%s%s%s" % (
            value[0:4], self.separator,
            value[4:8], self.separator,
            value[8:12]) \
            if self.word_sz == 4 else \
            "%s%s%s%s%s%s%s%s%s%s%s" % (
                value[0:2], self.separator,
                value[2:4], self.separator,
                value[4:6], self.separator,
                value[6:8], self.separator,
                value[8:10], self.separator,
                value[10:12]
            )

    def _value_handler(self, value):
        if value is None:
            return None if self.nullable else self._format_value('000000000000')
        return self._format_value(value)


class IPRequesties(ColumnAttributeAbstract):
    def __init__(self, **kwargs):
        kwargs['nullable'] = kwargs.get('nullable', True)
        kwargs['default'] = kwargs.get('default', None)
        kwargs['cast_write'] = self._cast_write
        kwargs['cast_read'] = self._cast_read
        kwargs['cast_order_by'] = None
        kwargs['cast_group_by'] = self._cast_read
        kwargs['strict_find'] = True
        self.default_v4mask = kwargs.pop('default_v4mask', 32)
        self.default_v6prefix = kwargs.pop('default_v6prefix', 64)
        # self.hide_host_mask = kwargs.pop('hide_host_mask', True)
        # TODO! '_cast_read()' with 'hide_host_mask' parameter (for MySQL implementation)
        self.hide_host_mask = False
        super(IPRequesties, self).__init__(**kwargs)
        self.enumerable_value = True
        self.value_class = ValueIPRequesties

    @staticmethod
    def _cast_write(context, pn):
        value = context.params[pn]
        if isinstance(value, str):
            return "%(" + pn + ")s"
        if value is None or isinstance(value, AttributeValueNull):
            return "NULL"
        if isinstance(value, ValueIPRequesties) and value.null:
            return "NULL"
        if context.session.dbms_type == 'mysql':
            return concat_(inet6_aton_(value.ipaddr), unhex_("%X" % int(value.ipmask)))
        context.params[pn] = value.string
        return "%(" + pn + ")s"

    def _cast_read(self, context):
        if context.session.dbms_type == 'mysql':
            return if_(
                length_(uncastable_(self)) == 17,
                concat_(
                    inet6_ntoa_(substr_(uncastable_(self), 1, 16)),
                    '/',
                    conv_(hex_(substr_(uncastable_(self), 17, 1)), 16, 10)
                ),
                concat_(
                    inet6_ntoa_(substr_(uncastable_(self), 1, 4)),
                    '/',
                    conv_(hex_(substr_(uncastable_(self), 5, 1)), 16, 10)
                )
            )
        elif context.session.dbms_type == 'postgresql':
            if self.hide_host_mask:
                return self
            else:
                return pg_text_(uncastable_(self))
        return self

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() != 'varbinary':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if struct['character_maximum_length'] != 64:
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['data_type']).lower() != 'inet':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        if session.dbms_type == 'mysql':
            d.append("VARBINARY(64)")
        elif session.dbms_type == 'postgresql':
            d.append("INET")
        else:
            raise TypeError("this column type is not supported for this database and has no alternative!")
        if not self.nullable:
            d.append('NOT NULL')
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            d.append("SET DATA TYPE INET")
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)


class GeometryAbstract(String):
    def __init__(self, **kwargs):
        kwargs['strict_find'] = True
        self.decimals = kwargs.pop('decimals', 0)
        super(GeometryAbstract, self).__init__(**kwargs)
        self.enumerable_value = True


class Dimensions2D(GeometryAbstract):
    def __init__(self, **kwargs):
        super(Dimensions2D, self).__init__(**kwargs)
        self.value_class = ValueDimensions2D


class Dimensions3D(GeometryAbstract):
    def __init__(self, **kwargs):
        super(Dimensions3D, self).__init__(**kwargs)
        self.value_class = ValueDimensions3D


class Box(GeometryAbstract):
    def __init__(self, **kwargs):
        super(Box, self).__init__(**kwargs)
        self.value_class = ValueBox


class Point(GeometryAbstract):
    def __init__(self, **kwargs):
        super(Point, self).__init__(**kwargs)
        self.value_class = ValuePoint


class Vector(GeometryAbstract):
    def __init__(self, **kwargs):
        super(Vector, self).__init__(**kwargs)
        self.value_class = ValueVector


class GeoCoordinate(GeometryAbstract):
    def __init__(self, **kwargs):
        super(GeoCoordinate, self).__init__(**kwargs)
        self.value_class = ValueGeoCoordinate


class IntegerChoice(Integer):
    def __init__(self, options, **kwargs):
        if not isinstance(options, dict):
            raise OrmParameterError("IntegerChoice requires 'options' argument to be a type (dict)!")
        if 'default' not in kwargs:
            raise OrmParameterError("IntegerChoice requires 'default' to be set!")
        kwargs['options'] = options
        super(IntegerChoice, self).__init__(**kwargs)

    def verify(self, value):
        if value not in self.options:
            raise OrmValueVerificationError("Unacceptable choice")
        value = super(IntegerChoice, self).verify(value)
        if value is None:
            return None
        return value


class StringChoice(String):
    def __init__(self, options, **kwargs):
        if not isinstance(options, dict):
            raise OrmParameterError("StringChoice requires 'options' argument to be a type (dict)!")
        if 'default' not in kwargs:
            raise OrmParameterError("StringChoice requires 'default' to be set!")
        kwargs['options'] = options
        super(StringChoice, self).__init__(**kwargs)

    def verify(self, value):
        if value not in self.options:
            raise OrmValueVerificationError("Unacceptable choice")
        value = super(StringChoice, self).verify(value)
        if value is None:
            return None
        return value





import datetime
from ..exceptions import OrmTypeError, OrmParameterError
from ..manage import manager
from ..mapper import mapper
from ..utils import gettext
from .base import DataAttributeAbstract, VirtualAttributeAbstract
from .derived import Integer, BigInteger, Float, BigFloat, Date, Time, DateTime, String, CaptionString, \
    UuidPrimaryKey, Dimensions2D, ColumnTextAbstract
from ..values import ValueInterval, ValueList, ValueDict, ValueMatrix, ValueFileAttachment, ValuePictureAttachment, \
    ValueFiles, ValuePictures, ValueSet, ValueInlineSet


ATTACH_BY_MOVE = 'move'
ATTACH_BY_LINK = 'link'
ATTACH_BY_COPY = 'copy'


class Interval(VirtualAttributeAbstract):
    def __init__(self, low_attr, high_attr, min_value=None, max_value=None, **kwargs):
        super(Interval, self).__init__(**kwargs)
        self.virtual = True
        self.enumerable_value = True
        self.min_value = min_value
        self.max_value = max_value
        self._low_attr = low_attr
        self._high_attr = high_attr
        self.k_low = None
        self.k_high = None
        self.childs_type = None

    def on_declare(self):
        if not isinstance(self._low_attr, DataAttributeAbstract) \
                or not isinstance(self._high_attr, DataAttributeAbstract):
            raise OrmTypeError(
                "Interval attribute requires first and second parameters to be a type of data attribute!"
            )
        self.k_low = self._low_attr.model_k
        self.k_high = self._high_attr.model_k
        if self._low_attr.data_type != self._high_attr.data_type:
            raise OrmTypeError(
                "Interval attribute requires that both low and high attributes be the same type!"
            )
        if self._low_attr.data_type not in ('int', 'float', 'date', 'time', 'datetime') \
                or self._high_attr.data_type not in ('int', 'float', 'date', 'time', 'datetime'):
            raise OrmTypeError(
                "Interval attribute supports only numeric or date/time values!"
            )
        self.childs_type = self._low_attr.data_type
        self.attributes = self.k_low, self.k_high

    def initialize(self, instance, k):
        virtual_value = ValueInterval()
        virtual_value.model_k = k
        virtual_value.instance = instance
        virtual_value.c = self
        instance.__setattrabs__(k, virtual_value)


class TypedIntervalAbstract(Interval):
    def __init__(self, c_type, min_value=None, max_value=None, k_low=None, k_high=None, **kwargs):
        super(TypedIntervalAbstract, self).__init__(None, None, min_value=min_value, max_value=max_value, **kwargs)
        self.childs_attr_type = c_type
        self.k_low = k_low
        self.k_high = k_high

    def on_declare(self):
        k_low = self.k_low or "%s__low" % self.model_k
        k_high = self.k_high or "%s__high" % self.model_k
        while hasattr(self.model, k_low) or hasattr(self.model, k_high):
            k_low += '_'
            k_high += '_'
        c_low = self.childs_attr_type(nullable=self.nullable)
        c_high = self.childs_attr_type(nullable=self.nullable)
        setattr(self.model, k_low, c_low)
        setattr(self.model, k_high, c_high)
        self._low_attr = c_low
        self._high_attr = c_high
        self.k_low = k_low
        self.k_high = k_high
        super(TypedIntervalAbstract, self).on_declare()


class IntInterval(TypedIntervalAbstract):
    def __init__(self, min_value=None, max_value=None, k_low=None, k_high=None, **kwargs):
        super(IntInterval, self).__init__(Integer, min_value, max_value, k_low, k_high, **kwargs)


class FloatInterval(TypedIntervalAbstract):
    def __init__(self, min_value=None, max_value=None, k_low=None, k_high=None, **kwargs):
        super(FloatInterval, self).__init__(Float, min_value, max_value, k_low, k_high, **kwargs)


class BigIntInterval(TypedIntervalAbstract):
    def __init__(self, min_value=None, max_value=None, k_low=None, k_high=None, **kwargs):
        super(BigIntInterval, self).__init__(BigInteger, min_value, max_value, k_low, k_high, **kwargs)


class BigFloatInterval(TypedIntervalAbstract):
    def __init__(self, min_value=None, max_value=None, k_low=None, k_high=None, **kwargs):
        super(BigFloatInterval, self).__init__(BigFloat, min_value, max_value, k_low, k_high, **kwargs)


class DateInterval(TypedIntervalAbstract):
    def __init__(self, min_value=None, max_value=None, k_low=None, k_high=None, **kwargs):
        super(DateInterval, self).__init__(Date, min_value, max_value, k_low, k_high, **kwargs)


class TimeInterval(TypedIntervalAbstract):
    def __init__(self, min_value=None, max_value=None, k_low=None, k_high=None, **kwargs):
        super(TimeInterval, self).__init__(Time, min_value, max_value, k_low, k_high, **kwargs)


class DateTimeInterval(TypedIntervalAbstract):
    def __init__(self, min_value=None, max_value=None, k_low=None, k_high=None, **kwargs):
        super(DateTimeInterval, self).__init__(DateTime, min_value, max_value, k_low, k_high, **kwargs)


class ArrayAbstract(VirtualAttributeAbstract):
    def __init__(self, **kwargs):
        super(ArrayAbstract, self).__init__(**kwargs)
        self.virtual = True
        self.enumerable_value = True
        self.value_class = None
        self.value_attr = None
        self.dict_key = None
        self.columns = None
        self.unique = False
        self.array_model = None
        self.in_composite_results = False

    def on_model_declared(self):
        child_attrs = dict()
        if self.dict_key is not None:
            child_attrs['_k_'] = self.dict_key
        for k in self.columns:
            child_attrs[k] = self.columns[k]
        self.array_model = mapper.define_attribute_model(self.model, self.model_k, child_attrs)

    def initialize(self, instance, k):
        virtual_value = self.value_class()
        virtual_value.model_k = k
        virtual_value.instance = instance
        virtual_value.c = self
        instance.__setattrabs__(k, virtual_value)

    def write(self, instance, key):
        virtual_value = getattr(instance, key)
        virtual_value.write()

    def log_repr(self, src, dst, instance=None):
        return gettext('Changed'), None


class List(ArrayAbstract):
    def __init__(self, column_attr, **kwargs):
        self.unique = bool(kwargs.pop('unique', False))
        super(List, self).__init__(**kwargs)
        self.value_class = ValueList
        self.columns = {'value': column_attr, 'pos': Integer(nullable=False, default=0)}


class Dict(ArrayAbstract):
    def __init__(self, key_attr, column_attr, **kwargs):
        super(Dict, self).__init__(**kwargs)
        self.value_class = ValueDict
        self.dict_key = key_attr
        self.columns = {'_v_': column_attr}


class Matrix(ArrayAbstract):
    def __init__(self, key_attr, columns, **kwargs):
        super(Matrix, self).__init__(**kwargs)
        self.value_class = ValueMatrix
        self.dict_key = key_attr
        self.columns = dict()
        if not columns or not isinstance(columns, dict):
            raise OrmParameterError(
                "Matrix attribute type requires 'columns' parameter to be set and be a type of (dict) with format"
                " {'<column_name>': <column_type>, ...}!"
            )
        self.columns = columns


class Set(ArrayAbstract):
    def __init__(self, options, length=255, **kwargs):
        kwargs['options'] = options
        super(Set, self).__init__(**kwargs)
        if not isinstance(options, dict):
            raise OrmParameterError("Set requires 'options' argument to be a type (dict)!")
        self.value_class = ValueSet
        self.columns = {'value': String(length, nullable=True, default=None)}


class InlineSet(ColumnTextAbstract):
    def __init__(self, options, **kwargs):
        kwargs['options'] = options
        kwargs['nullable'] = False
        super(InlineSet, self).__init__(**kwargs)
        self.enumerable_value = True
        self.value_class = ValueInlineSet

    def log_repr(self, src, dst, instance=None):
        return gettext('Changed'), None


class AttachmentAbstract(String):
    def __init__(self, **kwargs):
        kwargs['nullable'] = True
        kwargs['default'] = None
        self.storage = kwargs.pop('storage', None)
        self.tag = kwargs.pop('tag', False)
        self.push_rule = kwargs.pop('push_rule', ATTACH_BY_COPY)
        self.pull_rule = kwargs.pop('pull_rule', ATTACH_BY_COPY)
        self.check_existance = kwargs.pop('check_existance', True)
        if self.push_rule not in (ATTACH_BY_COPY, ATTACH_BY_LINK, ATTACH_BY_MOVE):
            raise OrmParameterError(
                "attachment rule can be one of ATTACH_BY_COPY, ATTACH_BY_LINK or ATTACH_BY_MOVE constant!"
            )
        if self.pull_rule not in (ATTACH_BY_COPY, ATTACH_BY_LINK, ATTACH_BY_MOVE):
            raise OrmParameterError(
                "attachment rule can be one of ATTACH_BY_COPY, ATTACH_BY_LINK or ATTACH_BY_MOVE constant!"
            )
        max_filename_length = kwargs.pop('max_filename_length', 1023)
        super(AttachmentAbstract, self).__init__(max_filename_length, **kwargs)
        self.in_composite_results = False

    def on_instance_load(self, instance):
        v = instance.__getattrabs__(self.model_k)
        v.on_instance_load()

    def on_after_delete(self, instance, hard=None):
        """ Deletes the attachment, if any attached, when instance been deleted. Not affecting when
        soft-deletion used. """
        if not hard:
            return
        iv = self.get_instance_value(instance)
        if not iv.is_attached:
            return
        iv.after_instance_deleted()

    def on_before_write(self, instance):
        if not instance.__existing__:
            return
        iv = getattr(instance, self.model_k)
        iv.before_instance_write()

    def on_after_write(self, instance):
        iv = getattr(instance, self.model_k)
        before_ = iv.filename
        iv.after_instance_write()
        after_ = iv.filename
        if before_ == after_:
            return
        # Need to update this instance because the file name been changed
        # after physical attaching to the storage and we need to store correct
        # value in the database instead of initial one.
        instance.write(attrs=[self.model_k, ])

    def push_file(self, *args, **kwargs):
        """ Virtual. Real method is defined in the 'ValueFileAttachment' class. """
        pass

    def pull_file(self, *args, **kwargs):
        """ Virtual. Real method is defined in the 'ValueFileAttachment' class. """
        pass

    def remove_file(self, *args, **kwargs):
        """ Virtual. Real method is defined in the 'ValueFileAttachment' class. """
        pass

    @property
    def abs_path(self):
        """ Virtual. Real method is defined in the 'ValueFileAttachment' class. """
        return None

    @property
    def rel_path(self):
        """ Virtual. Real method is defined in the 'ValueFileAttachment' class. """
        return None

    @property
    def is_attached(self):
        """ Virtual. Real method is defined in the 'ValueFileAttachment' class. """
        return None


class FileAttachment(AttachmentAbstract):
    def __init__(self, **kwargs):
        super(FileAttachment, self).__init__(**kwargs)
        self.value_class = ValueFileAttachment

    def on_declare(self):
        if self.tag is False:
            self.tag = self.model.__meta__.tag
        if self.storage is None:
            self.storage = manager.get_default_file_storage(tag=self.tag)


class PictureAttachment(AttachmentAbstract):
    def __init__(self, **kwargs):
        self.max_width = kwargs.pop('max_width', None)
        self.min_width = kwargs.pop('min_width', None)
        self.max_height = kwargs.pop('max_height', None)
        self.min_height = kwargs.pop('min_height', None)
        self.image_format = kwargs.pop('image_format', 'JPEG')
        super(PictureAttachment, self).__init__(**kwargs)
        self.value_class = ValuePictureAttachment

    def on_declare(self):
        if self.tag is False:
            self.tag = self.model.__meta__.tag
        if self.storage is None:
            self.storage = manager.get_default_picture_storage(tag=self.tag)

    def push_image(self, *args, **kwargs):
        """ Virtual. Real method is defined in the 'ValuePictureAttachment' class. """
        pass

    def push_image_to(self, *args, **kwargs):
        """ Virtual. Real method is defined in the 'ValuePictureAttachment' class. """
        pass

    def pull_image_to(self, *args, **kwargs):
        """ Virtual. Real method is defined in the 'ValuePictureAttachment' class. """
        pass

    def pull_image(self, *args, **kwargs):
        """ Virtual. Real method is defined in the 'ValuePictureAttachment' class. """
        pass

    def remove_image(self, *args, **kwargs):
        """ Virtual. Real method is defined in the 'ValuePictureAttachment' class. """
        pass


class FilesAbstract(VirtualAttributeAbstract):
    def __init__(self, **kwargs):
        self.storage = kwargs.pop('storage', None)
        self.tag = kwargs.pop('tag', False)
        self.push_rule = kwargs.pop('push_rule', ATTACH_BY_COPY)
        self.pull_rule = kwargs.pop('pull_rule', ATTACH_BY_COPY)
        if self.push_rule not in (ATTACH_BY_COPY, ATTACH_BY_LINK, ATTACH_BY_MOVE):
            raise OrmParameterError(
                "attachment rule can be one of ATTACH_BY_COPY, ATTACH_BY_LINK or ATTACH_BY_MOVE constant!"
            )
        if self.pull_rule not in (ATTACH_BY_COPY, ATTACH_BY_LINK, ATTACH_BY_MOVE):
            raise OrmParameterError(
                "attachment rule can be one of ATTACH_BY_COPY, ATTACH_BY_LINK or ATTACH_BY_MOVE constant!"
            )
        self.max_filename_length = kwargs.pop('max_filename_length', 1023)
        self.extra_attrs = kwargs.pop('extra_attrs', None)
        super(FilesAbstract, self).__init__(**kwargs)
        self.virtual = True
        self.enumerable_value = True
        self.in_composite_results = False
        self.value_class = None
        self.array_model = None
        self.array_model_attrs = {
            'id': UuidPrimaryKey(),
            'time_created': DateTime(nullable=False, default=datetime.datetime.now, caption=gettext('Created')),
            'title': CaptionString(caption=gettext('Name')),
            'filesize': Integer(nullable=False, default=0, caption=gettext('Size'))
        }
        self.array_model_attrs['id'].declaration_order = 0
        if self.extra_attrs is not None and not isinstance(self.extra_attrs, dict):
            raise OrmParameterError("Files|Pictures attribute's 'extra_attrs' must be of type (dict)!")
        elif self.extra_attrs is not None:
            for k in self.extra_attrs:
                self.array_model_attrs[k] = self.extra_attrs[k]

    def __setattr__(self, key, value):
        if key in ('push_rule', 'pull_rule', 'max_filename_length', 'max_width', 'min_width', 'max_height',
                   'min_height', 'image_format', 'storage'):
            array_model = getattr(self, 'array_model', None)
            if array_model is not None:
                setattr(array_model.attachment, key, value)
        super(FilesAbstract, self).__setattr__(key, value)

    def on_model_declared(self):
        self.array_model = mapper.define_attribute_model(self.model, self.model_k, self.array_model_attrs)

    def initialize(self, instance, k):
        virtual_value = self.value_class()
        virtual_value.model_k = k
        virtual_value.instance = instance
        virtual_value.c = self
        instance.__setattrabs__(k, virtual_value)

    def write(self, instance, key):
        virtual_value = getattr(instance, key)
        virtual_value.write()

    def on_after_delete(self, instance, hard=None):
        """ Deletes the attachment, if any attached, when instance been deleted. Not affecting when
        soft-deletion used. """
        if not hard:
            return
        iv = self.get_instance_value(instance)
        if not iv.is_attached:
            return
        iv.after_parent_instance_deleted()

    def log_repr(self, src, dst, instance=None):
        return gettext('Changed'), None


class Files(FilesAbstract):
    def __init__(self, **kwargs):
        super(Files, self).__init__(**kwargs)
        self.value_class = ValueFiles
        self.array_model_attrs['attachment'] = FileAttachment(caption=gettext('File'),
                                                              storage=self.storage,
                                                              tag=self.tag,
                                                              push_rule=self.push_rule,
                                                              pull_rule=self.pull_rule,
                                                              max_filename_length=self.max_filename_length,
                                                              check_existance=False)


class Pictures(FilesAbstract):
    def __init__(self, **kwargs):
        self.max_width = kwargs.pop('max_width', None)
        self.min_width = kwargs.pop('min_width', None)
        self.max_height = kwargs.pop('max_height', None)
        self.min_height = kwargs.pop('min_height', None)
        self.image_format = kwargs.pop('image_format', None)
        super(Pictures, self).__init__(**kwargs)
        self.value_class = ValuePictures
        self.array_model_attrs['attachment'] = PictureAttachment(caption=gettext('Picture'),
                                                                 storage=self.storage,
                                                                 tag=self.tag,
                                                                 max_filename_length=self.max_filename_length,
                                                                 max_width=self.max_width,
                                                                 min_width=self.min_width,
                                                                 max_height=self.max_height,
                                                                 min_height=self.min_height,
                                                                 image_format=self.image_format,
                                                                 check_existance=False)
        self.array_model_attrs['dimensions'] = Dimensions2D(nullable=False, caption='Dimensions')
        self.array_model_attrs['format'] = String(16, nullable=False, default='', caption=gettext('Format'))




